@@ -2,11 +2,11 @@
 
                           Firewall Builder
 
-                 Copyright (C) 2000 Vadim Kurland, Vadim Zaliva
+                 Copyright (C) 2000 NetCitadel, LLC
 
   Author:  Vadim Zaliva     lord@crocodile.org
 
-  $Id: Constants.cc,v 1.4 2001/12/19 12:53:59 lord Exp $
+  $Id: Constants.cc,v 1.6 2002/09/12 03:04:09 vkurland Exp $
 
 
   This program is free software which we release under the GNU General Public
@@ -30,6 +30,11 @@
 using namespace std;
 using namespace libfwbuilder;
 
+const string Constants::getLibraryDescription()
+{
+    return "Firewall Builder API ";
+}
+
 const string Constants::getLibraryVersion() 
 { 
     return string(LIBFWBUILDER_FORMAT_VERSION); 
