@@ -2,11 +2,11 @@
 
                           Firewall Builder
 
-                 Copyright (C) 2000 Vadim Kurland
+                 Copyright (C) 2000 NetCitadel, LLC
 
   Author:  Vadim Kurland     vadim@vk.crocodile.org
 
-  $Id: FWException.cc,v 1.3 2001/12/19 12:46:03 lord Exp $
+  $Id: FWException.cc,v 1.5 2002/08/29 00:34:09 vkurland Exp $
 
 
   This program is free software which we release under the GNU General Public
@@ -34,7 +34,7 @@ FWException::FWException(const string &r)
     reason = r;
 }
 
-const string &FWException::toString()
+const string &FWException::toString() const
 {
     return reason;
 }
