@@ -2,11 +2,11 @@
 
                           Firewall Builder
 
-                 Copyright (C) 2000 Vadim Kurland
+                 Copyright (C) 2000 NetCitadel, LLC
 
   Author:  Vadim Kurland     vadim@vk.crocodile.org
 
-  $Id: FWIntervalReference.cc,v 1.3 2001/12/19 12:46:03 lord Exp $
+  $Id: FWIntervalReference.cc,v 1.6 2002/08/29 00:34:09 vkurland Exp $
 
 
   This program is free software which we release under the GNU General Public
@@ -32,15 +32,20 @@ using namespace libfwbuilder;
 
 const char *FWIntervalReference::TYPENAME={"IntervalRef"};
 
-FWIntervalReference::FWIntervalReference(FWObject *p):FWReference(p)
+
+//FWIntervalReference::FWIntervalReference(Interval *p):FWReference(p){}
+//FWIntervalReference::FWIntervalReference(IntervalGroup *p):FWReference(p){}
+
+FWIntervalReference::FWIntervalReference():FWReference(){}
+FWIntervalReference::FWIntervalReference(const FWObject *root) : FWReference(root) {} 
+
+void FWIntervalReference::setPointer(Interval *p)
 {
+    //TODO: add check if it is right type.
+    FWReference::setPointer(p);
 }
 
-FWIntervalReference::FWIntervalReference():FWReference()
-{
-}
-
-void FWIntervalReference::setPointer(FWObject *p)
+void FWIntervalReference::setPointer(IntervalGroup *p)
 {
     //TODO: add check if it is right type.
     FWReference::setPointer(p);
