@@ -2,11 +2,11 @@
 
                           Firewall Builder
 
-                 Copyright (C) 2000 Vadim Kurland
+                 Copyright (C) 2000 NetCitadel, LLC
 
   Author:  Vadim Kurland     vadim@vk.crocodile.org
 
-  $Id: FWObject.cc,v 1.8 2001/12/28 08:49:18 lord Exp $
+  $Id: FWObject.cc,v 1.32 2003/03/02 21:47:29 vkurland Exp $
 
 
   This program is free software which we release under the GNU General Public
@@ -40,6 +40,8 @@
 #include <functional>
 #include <stack>
 
+#include <assert.h>
+
 using namespace std;
 using namespace libfwbuilder;
 
@@ -47,24 +49,39 @@ const char *FWObject::TYPENAME={"UNDEF"};
 string FWObject::NOT_FOUND="";
 
 
-void FWObject::fromXML(xmlNodePtr root)
+void FWObject::fromXML(xmlNodePtr root) throw(FWException)
 {
     assert(root!=NULL);
 
     const char *n;
 
     n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("name")));
-    if(n)  setName(n);
+    if(n)
+    {
+        setName(n);
+        FREEXMLBUFF(n);
+    }
 
     n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("id")));
-    if(n)  setId(n);
+    if(n)
+    {
+        setId(n);
+        FREEXMLBUFF(n);
+    }
 
     n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("comment")));
-    if(n)  setComment(XMLTools::unquote_linefeeds(n));
+    if(n)
+    {
+        setComment(XMLTools::unquote_linefeeds(n));
+        FREEXMLBUFF(n);
+    }
 
     n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("library")));
-    if(n)  setLibrary(n);
-
+    if(n)
+    {
+        setLibrary(n);
+        FREEXMLBUFF(n);
+    }
 
     ref_counter=0;
 
@@ -76,7 +93,16 @@ void FWObject::fromXML(xmlNodePtr root)
             if(o!=NULL) 
             {
 		add(o);
-		o->fromXML(cur);
+                try
+                {
+                    o->fromXML(cur);
+                } catch(FWException &ex)
+                {
+                    map<string, string> &properties = ex.getProperties();
+                    if(properties.find("failed_element")==properties.end())
+                        properties["failed_element"]=o->getTypeName();
+                    throw;
+                }
 	    }
         }
     }
@@ -84,19 +110,19 @@ void FWObject::fromXML(xmlNodePtr root)
     setDirty(false);
 }
 
-xmlNodePtr FWObject::toXML(xmlNodePtr xml_parent_node)
+xmlNodePtr FWObject::toXML(xmlNodePtr xml_parent_node) throw(FWException)
 {
     return toXML(xml_parent_node, true);
 }
 
-xmlNodePtr FWObject::toXML(xmlNodePtr parent, bool process_children)
+xmlNodePtr FWObject::toXML(xmlNodePtr parent, bool process_children) throw(FWException)
 {
     xmlNodePtr me = xmlNewChild(parent, NULL, xml_name.empty()?STRTOXMLCAST(getTypeName()):STRTOXMLCAST(xml_name), NULL);
 
-    for(map<string, string>::iterator i=data.begin(); i!=data.end(); ++i) 
+    for(map<string, string>::const_iterator i=data.begin(); i!=data.end(); ++i) 
     {
-        string name  = (*i).first;
-        string value = (*i).second;
+        const string &name  = (*i).first;
+        const string &value = (*i).second;
 
         xmlAttrPtr pr = xmlNewProp(me, 
                                    STRTOXMLCAST(name) , 
@@ -112,11 +138,8 @@ xmlNodePtr FWObject::toXML(xmlNodePtr parent, bool process_children)
     }
 
     if(process_children)
-    {
-        vector<FWObject*>::iterator j;
-        for(j=begin(); j!=end(); ++j) 
+        for(list<FWObject*>::const_iterator j=begin(); j!=end(); ++j) 
             (*j)->toXML(me);
-    }
     
     return me;
 }
@@ -124,6 +147,8 @@ xmlNodePtr FWObject::toXML(xmlNodePtr parent, bool process_children)
 
 FWObject::FWObject()
 {
+    setName("New object");
+    
     ref_counter = 0;
     parent      = NULL;
 
@@ -139,12 +164,17 @@ FWObject::FWObject(const FWObject &c)
 }
 
 
-FWObject::FWObject(FWObject *par)
+/**
+ * we do not need to do anything with tree root in FWObject, but
+ * derived classes might want to use it to create children objects and
+ * add references etc.
+ */
+FWObject::FWObject(const FWObject *root)
 {
     setName("New object");
     
     ref_counter = 0    ;
-    parent      = par  ;
+    parent      = NULL ;
     
     // When object created we assign it unique Id
     setId(FWObjectDatabase::db->generateUniqueId());
@@ -152,14 +182,13 @@ FWObject::FWObject(FWObject *par)
     setDirty(false);
 }
 
-
 FWObject::~FWObject() 
 {
     clearChildren();
     data.clear();
 }
 
-FWObject* FWObject::getParent()
+FWObject* FWObject::getParent() const
 {
     return parent;
 }
@@ -174,56 +203,51 @@ void FWObject::setXMLName(const string &n)
     xml_name = n;
 }
 
-FWObject* FWObject::_find(const string& name)   // find child by name
+FWObject* FWObject::_find(const string& name) const
 {
-    iterator i=find_if(begin(),end(), FWObjectNameEQPredicate(name));
+    const_iterator i=find_if(begin(),end(), FWObjectNameEQPredicate(name));
     return i==end()?NULL:(*i);
 }
 
-FWObject& FWObject::operator=(const FWObject &x)
+FWObject& FWObject::operator=(const FWObject &x)  throw(FWException)
 {
-    clearChildren();
+    return duplicate(&x, FALSE);
+}
 
-    ref_counter = 0          ;
-    parent      = x.parent   ;
-    xml_name    = x.xml_name ;
-    data        = x.data     ;
-    
-    for(vector<FWObject*>::const_iterator m=x.begin(); m!=x.end(); ++m) 
+FWObject& FWObject::duplicate(const FWObject *x, bool preserve_id) throw(FWException)
+{
+    shallowDuplicate(x,preserve_id);
+
+    destroyChildren();
+
+    for(list<FWObject*>::const_iterator m=x->begin(); m!=x->end(); ++m) 
     {
         FWObject *o=*m;
-        FWObject *o1 = FWObjectDatabase::db->create(o->getTypeName());
-        assert(o1!=NULL);
-        *o1=*o;
+        FWObject *o1=FWObjectDatabase::db->create(o->getTypeName());
+        if(!o1)
+            throw FWException(string("Error creating object with type: ")+o->getTypeName());
+        o1->duplicate(o, preserve_id);
         add(o1);
     }
-    setDirty(true);
 
+    setDirty(true);
     return *this;
 }
 
-FWObject& FWObject::duplicate(const FWObject *x)
+FWObject& FWObject::shallowDuplicate(const FWObject *x, bool preserve_id) throw(FWException)
 {
-    string id=getId();
-    
-    clearChildren();
+    string id = getId();
 
-    data=x->data;
+    data = x->data;
 
-    if(id!="")   // some objects do not have ID per DTD (e.g. Src, Dst, etc.)
+    if(!preserve_id)
+    {
+        ref_counter = 0           ;
+        xml_name    = x->xml_name ;
+    } else if(id!="")   // some objects do not have ID per DTD (e.g. Src, Dst, etc.)
 	setId(id);
 
-    for(vector<FWObject*>::const_iterator m=x->begin(); m!=x->end(); ++m) 
-    {
-        FWObject *o=*m;
-        FWObject *o1=FWObjectDatabase::db->create( o->getTypeName());
-        assert(o1!=NULL);
-        o1->duplicate(o);
-        add(o1);
-    }
-    
     setDirty(true);
-
     return *this;
 }
 
@@ -384,7 +408,7 @@ void   FWObject::dump(std::ostream &f,bool recursive,bool brief,int offset)
 	f << endl;
 
 	if (recursive) {
-	    vector<FWObject*>::iterator m;
+	    list<FWObject*>::iterator m;
 	    for (m=begin(); m!=end(); ++m) {
 		if (  (o=(*m))!=NULL)  o->dump(f,recursive,brief,offset+2);
 	    }
@@ -412,7 +436,7 @@ void   FWObject::dump(std::ostream &f,bool recursive,bool brief,int offset)
 	    f << (*d).first << ": " << (*d).second << endl;
 	}
 	if (recursive) {
-	    vector<FWObject*>::iterator m;
+	    list<FWObject*>::iterator m;
 	    for (m=begin(); m!=end(); ++m) {
 		if (  (o=(*m))!=NULL)  o->dump(f,recursive,brief,offset+2);
 	    }
@@ -434,9 +458,9 @@ void FWObject::addAt(const string& where_id, FWObject *obj)
     p->add(obj);
 }
 
-void FWObject::add(FWObject *obj)
+void FWObject::add(FWObject *obj,bool validate)
 {
-    if(validateChild(obj)) 
+    if(!validate || validateChild(obj)) 
     {
 	push_back(obj);
 	_adopt(obj);
@@ -444,9 +468,11 @@ void FWObject::add(FWObject *obj)
     }
 }
 
-FWReference* FWObject::createRef( FWObject* obj )
+FWReference* FWObject::createRef()
 {
-    return new FWObjectReference(obj);
+    FWObjectReference *ref=new FWObjectReference();
+    ref->setPointer(this);
+    return ref;
 }
 
 
@@ -454,7 +480,7 @@ void FWObject::addRef(FWObject *obj)
 {
     if(validateChild(obj)) 
     {
-	FWReference *oref = createRef(obj);
+	FWReference *oref = obj->createRef();
 	obj->ref();
 
 	push_back(oref);
@@ -468,7 +494,7 @@ void FWObject::insert_before(FWObject *o1, FWObject *obj)
     if(!obj) 
         return;
     
-    vector<FWObject*>::iterator m=find(begin(),end(),o1);
+    list<FWObject*>::iterator m=find(begin(),end(),o1);
     if(m!=end())
     {
         insert(m, obj);
@@ -482,7 +508,7 @@ void FWObject::insert_after(FWObject *o1, FWObject *obj)
     if(!obj) 
         return;
 
-    vector<FWObject*>::iterator m=find(begin(),end(),o1);
+    list<FWObject*>::iterator m=find(begin(),end(),o1);
     if(m!=end())
     {
         insert(++m,obj);
@@ -493,7 +519,7 @@ void FWObject::insert_after(FWObject *o1, FWObject *obj)
 
 void FWObject::swapObjects(FWObject *o1, FWObject *o2)
 {
-    for(vector<FWObject*>::iterator m=begin(); m!=end(); ++m) 
+    for(list<FWObject*>::iterator m=begin(); m!=end(); ++m) 
     {
         if(*m==o1) 
         {
@@ -506,7 +532,7 @@ void FWObject::swapObjects(FWObject *o1, FWObject *o2)
     setDirty(true);
 }
 
-void FWObject::remove(FWObject *obj, bool delete_if_last=true)
+void FWObject::remove(FWObject *obj, bool delete_if_last)
 {
     FWObject::iterator fi=std::find(begin(), end(), obj);
     if(fi!=end())
@@ -545,13 +571,14 @@ void FWObject::removeAllInstances(FWObject *rm)
 
 void FWObject::findAllReferences(const FWObject *obj, std::set<FWReference*> &res)
 {
-    for(vector<FWObject*>::iterator m=begin(); m!=end(); ++m) 
+    string obj_id=obj->getId();
+    for(list<FWObject*>::iterator m=begin(); m!=end(); ++m) 
     {
         FWObject *o=*m;
         FWReference *oref=FWReference::cast(o);
         if(oref)
         {
-	    if(oref->getPointer()==obj) 
+	    if(oref->getPointerId()==obj_id) 
                 res.insert(oref);
 	} else
         {
@@ -569,13 +596,14 @@ set<FWReference*> FWObject::findAllReferences(const FWObject *obj)
 
 void FWObject::removeRef(FWObject *obj)
 {
-    for(vector<FWObject*>::iterator m=begin(); m!=end(); ++m) 
+    string obj_id=obj->getId();
+    for(list<FWObject*>::iterator m=begin(); m!=end(); ++m) 
     {
         FWObject *o=*m;
         if(FWReference::cast(o)!=NULL)
         {
 	    FWReference *oref=FWReference::cast(o);
-	    if(oref && oref->getPointer()==obj)
+	    if(oref && oref->getPointerId()==obj_id)
             {
                 // do not delete object even if this reference was the last one (?)
 		obj->unref();  
@@ -622,11 +650,44 @@ bool FWObject::validateChild(FWObject *obj)
     return true;
 }
 
-void FWObject::clearChildren()
+void FWObject::destroyChildren()
 {
-    for(vector<FWObject*>::iterator m=begin(); m!=end(); ++m) 
+    for(list<FWObject*>::iterator m=begin(); m!=end(); ++m) 
     {
         FWObject *o=*m;
+        o->destroyChildren();
+        delete o;
+    }
+    clear();
+}
+
+/*
+ *  even if I run this method with flag recursive=true, it does not
+ *  guarantee that there will be no objects lost in the memory. If
+ *  some children of this are referenced from somewhere else, they
+ *  won't be deleted because their reference counter is >1. This is
+ *  bad because it leads to a situation when object is not a part of
+ *  the tree anymore, but reference to it does exist.
+ *
+ *  If this method is called with recursive=false, then it deletes
+ *  only immediate children of this, leaving their children hanging in
+ *  the memory. TODO: research whether we ever need to call it with
+ *  recursive=false
+ *
+ *  In other words, this method leaves tree in inconsistent state. At
+ *  this time I am just using it carefully, only when I copy objects
+ *  between main tree and scratch pad, and when I create copies of
+ *  objects. In both cases children of this will be immediately
+ *  restored after call to clearChildren.
+ *
+ *                                              05/08/02 vk
+ */
+void FWObject::clearChildren(bool recursive)
+{
+    for(list<FWObject*>::iterator m=begin(); m!=end(); ++m) 
+    {
+        FWObject *o=*m;
+        if (recursive) o->clearChildren(recursive);
         o->unref();
         if(o->ref_counter==0) 
             delete o;
@@ -635,77 +696,59 @@ void FWObject::clearChildren()
     setDirty(true);
 }
 
-void FWObject::clearChildren(const string &type_name)
-{
- loop:
-    for(vector<FWObject*>::iterator m=begin(); m!=end(); ++m) 
-    {
-        FWObject *o=(*m);
-        if(o->getTypeName()==type_name)
-        {
-	    remove(o); 
-            goto loop;
-        }
-    }
-}
-
 int FWObject::getChildrenCount()
 {
     return(size());
 }
 
-FWObject* FWObject::getRoot()
+FWObject* FWObject::getRoot() const
 {
-    FWObject *p=this;
+    const FWObject *p=this;
 
     while(p->getParent()!=NULL) p=p->getParent();
-    return p;
+    return (FWObject*)p;
 }
 
-
-FWObject* FWObject::getById  (const string &id, bool recursive,bool dereference)
+/*
+ * returns true if this is either direct child of obj, or a child
+ * of one of its children and so on.
+ */
+bool FWObject::isChildOf(FWObject *obj)
 {
-    if(id==getId())
-        return this;
+    if (this==obj) return false;
+    FWObject *p=this;
+    while (p!=NULL && p!=obj) p=p->getParent();
+    return (p==obj);
+}
+
+FWObject* FWObject::getById  (const string &id, bool recursive)
+{
+    if(id==getId())  return this;
     
-    vector<FWObject*>::iterator j;
+    list<FWObject*>::iterator j;
     for(j=begin(); j!=end(); ++j)     
     {
         FWObject *o=*j;
         string oid=o->getId();
-        if(id==oid) 
-            return o;
-        /*
-         *  if we follow references, do it only for ID comparison. Do not
-         *  follow references recursively as it poses danger of loops
-         */
-        if (dereference && FWReference::cast(o)!=NULL) {
-            FWObject *o1=FWReference::cast(o)->getPointer();
-            if (o1 && id==o1->getId() )
-                return o1;
-        }
+        if(id==oid) return o;
 
-        if(recursive)  
-        {
-            o=o->getById(id, true);
-            if(o) 
-                return o;
-        }
+        if(recursive && (o=o->getById(id, true))!=NULL ) return o;
     }
     return NULL; // not found
 }
 
 
-FWObject* FWObject::getFirstByType(const string &type_name)
+FWObject* FWObject::getFirstByType(const string &type_name) const
 {
-    iterator i=find_if(begin(),end(), FWObjectTypeNameEQPredicate(type_name));
+    const_iterator i=find_if(begin(),end(), FWObjectTypeNameEQPredicate(type_name));
     return i==end()?NULL:(*i);
 }
 
-vector<FWObject*> FWObject::getByType(const string &type_name)
+list<FWObject*> FWObject::getByType(const string &type_name) const
 {
-    vector<FWObject*> res;
-    for (iterator i=begin(); i!=end(); ++i) {
+    list<FWObject*> res;
+    for(const_iterator i=begin(); i!=end(); ++i) 
+    {
 	i=find_if( i, end(), FWObjectTypeNameEQPredicate(type_name));
 	if (i==end()) break;
 	res.push_back(*i);
@@ -713,11 +756,16 @@ vector<FWObject*> FWObject::getByType(const string &type_name)
     return res; 
 }
 
+FWObjectTypedChildIterator FWObject::findByType(const std::string &type_name) const
+{
+    return FWObjectTypedChildIterator(this, type_name);
+}
+
 void  FWObject::setDirty(bool f, bool recursive)
 {
     dirty=f;
     if(recursive) 
-	for(vector<FWObject*>::iterator m=begin(); m!=end(); ++m) 
+	for(list<FWObject*>::iterator m=begin(); m!=end(); ++m) 
             (*m)->setDirty(f, TRUE);
 }
 
@@ -728,7 +776,7 @@ bool FWObject::isDirty(bool recursive)
 
     if(recursive) 
     {
-	for(vector<FWObject*>::iterator m=begin(); m!=end(); ++m)
+	for(list<FWObject*>::iterator m=begin(); m!=end(); ++m)
             if((*m)->isDirty(recursive))
                 return TRUE;
     }
@@ -758,6 +806,60 @@ class sort_order_func_adaptor
 
 void FWObject::sortChildren()
 {
-    std::sort(begin(), end(), sort_order_func_adaptor(this));
+    this->sort(sort_order_func_adaptor(this));
     setDirty(true);
 }
+
+FWObjectTypedChildIterator::FWObjectTypedChildIterator(const FWObjectTypedChildIterator &o):
+    type_name(o.type_name),real_iterator(o.real_iterator), _begin(o._begin), _end(o._end)
+{
+}
+
+FWObjectTypedChildIterator::FWObjectTypedChildIterator(const FWObject *o, const std::string &_type_name)
+{
+    type_name     = _type_name ;
+    _end          = o->end()   ;
+    real_iterator = o->begin() ;
+    // position to first element
+    while(real_iterator!=_end && (*real_iterator)->getTypeName()!=type_name)
+        real_iterator++;
+    _begin = real_iterator;
+}
+
+bool FWObjectTypedChildIterator::operator==(const FWObject::const_iterator& __x) const 
+{ 
+    return real_iterator == __x; 
+}
+
+bool FWObjectTypedChildIterator::operator!=(const FWObject::const_iterator& __x) const 
+{ 
+    return real_iterator != __x; 
+}
+    
+FWObject *FWObjectTypedChildIterator::operator*() const 
+{ 
+    return *real_iterator;
+}
+
+FWObjectTypedChildIterator& FWObjectTypedChildIterator::operator++() 
+{
+    if(real_iterator==_end)
+        return *this;
+    do
+    {
+        real_iterator++;
+    } while(real_iterator!=_end && (*real_iterator)->getTypeName()!=type_name);
+        return *this;
+}
+
+FWObjectTypedChildIterator& FWObjectTypedChildIterator::operator--() 
+{ 
+    if(real_iterator==_begin)
+        return *this;
+    do
+    {
+        real_iterator--;
+    } while(real_iterator!=_begin && (real_iterator==_end || (*real_iterator)->getTypeName()!=type_name));
+    return *this;
+}
+
