@@ -2,11 +2,11 @@
 
                           Firewall Builder
 
-                 Copyright (C) 2000 Vadim Kurland
+                 Copyright (C) 2000 NetCitadel, LLC
 
   Author:  Vadim Kurland     vadim@vk.crocodile.org
 
-  $Id: FWObjectReference.cc,v 1.3 2001/12/19 12:46:12 lord Exp $
+  $Id: FWObjectReference.cc,v 1.5 2002/08/29 00:34:09 vkurland Exp $
 
 
   This program is free software which we release under the GNU General Public
@@ -26,26 +26,36 @@
 
 #include <fwbuilder/FWObjectReference.hh>
 #include <fwbuilder/FWObjectDatabase.hh>
+#include <fwbuilder/ObjectGroup.hh>
 #include <fwbuilder/XMLTools.hh>
 
 using namespace libfwbuilder;
 
 const char *FWObjectReference::TYPENAME={"ObjectRef"};
 
-FWObjectReference::FWObjectReference(FWObject *p):FWReference(p)
-{
-}
+//FWObjectReference::FWObjectReference(FWObject *p):FWReference(p)
+//{
+//}
 
 FWObjectReference::FWObjectReference():FWReference()
 {
 }
 
+FWObjectReference::FWObjectReference(const FWObject *root) : FWReference(root)
+{}
+
 void FWObjectReference::setPointer(FWObject *p)
 {
     //TODO: add check if it is right type.
     FWReference::setPointer(p);
 }
 
+void FWObjectReference::setPointer(ObjectGroup *p)
+{
+    //TODO: add check if it is right type.
+    FWReference::setPointer(p);
+}
+
 
 
 
