@@ -2,11 +2,11 @@
 
                           Firewall Builder
 
-                 Copyright (C) 2000 Vadim Kurland
+                 Copyright (C) 2000 NetCitadel, LLC
 
   Author:  Vadim Kurland     vadim@vk.crocodile.org
 
-  $Id: FWServiceReference.cc,v 1.3 2001/12/19 12:46:21 lord Exp $
+  $Id: FWServiceReference.cc,v 1.7 2002/08/29 00:34:09 vkurland Exp $
 
 
   This program is free software which we release under the GNU General Public
@@ -26,23 +26,29 @@
 
 #include <fwbuilder/FWServiceReference.hh>
 #include <fwbuilder/FWObjectDatabase.hh>
+#include <fwbuilder/Service.hh>
+#include <fwbuilder/ServiceGroup.hh>
 #include <fwbuilder/XMLTools.hh>
 
 using namespace libfwbuilder;
 
 const char *FWServiceReference::TYPENAME={"ServiceRef"};
 
-FWServiceReference::FWServiceReference(FWObject *p):FWReference(p)
+FWServiceReference::FWServiceReference():FWReference(){}
+FWServiceReference::FWServiceReference(const FWObject *root) : FWReference(root) {}
+
+
+//FWServiceReference::FWServiceReference(Service *p):FWReference(p){}
+//FWServiceReference::FWServiceReference(ServiceGroup *p):FWReference(p){}
+
+void FWServiceReference::setPointer(Service *p)
 {
+    FWReference::setPointer(p);
 }
 
-FWServiceReference::FWServiceReference():FWReference()
-{
-}
 
-void FWServiceReference::setPointer(FWObject *p)
+void FWServiceReference::setPointer(ServiceGroup *p)
 {
-    //TODO: add check if it is right type.
     FWReference::setPointer(p);
 }
 
