@@ -2,11 +2,11 @@
 
                           Firewall Builder
 
-                 Copyright (C) 2000 Vadim Kurland
+                 Copyright (C) 2000 NetCitadel, LLC
 
   Author:  Vadim Kurland     vadim@vk.crocodile.org
 
-  $Id: Group.cc,v 1.3 2001/12/19 12:46:26 lord Exp $
+  $Id: Group.cc,v 1.5 2002/08/29 00:34:09 vkurland Exp $
 
 
   This program is free software which we release under the GNU General Public
@@ -39,9 +39,8 @@ const char *Group::TYPENAME={"Group"};
 /*
  *  Group should inherit a list of allowed types from its parent upon creation
  */
-Group::Group()  
-{
-}
+Group::Group()  {}
+Group::Group(const FWObject *root) : FWObject(root) {}
 
 Group::~Group() 
 {
