@@ -2,11 +2,11 @@
 
                           Firewall Builder
 
-                 Copyright (C) 2000 Vadim Kurland
+                 Copyright (C) 2000 NetCitadel, LLC
 
   Author:  Vadim Kurland     vadim@vk.crocodile.org
 
-  $Id: ICMPService.cc,v 1.3 2001/12/19 12:46:38 lord Exp $
+  $Id: ICMPService.cc,v 1.13 2002/09/08 19:28:16 vkurland Exp $
 
 
   This program is free software which we release under the GNU General Public
@@ -24,27 +24,49 @@
 
 */
 
+#include <assert.h>
+
 
 #include <fwbuilder/ICMPService.hh>
 #include <fwbuilder/XMLTools.hh>
 
 using namespace libfwbuilder;
+using namespace std;
 
 const char *ICMPService::TYPENAME={"ICMPService"};
 
-ICMPService::ICMPService() {}
+ICMPService::ICMPService() 
+{
+    setStr("type", "-1");
+    setStr("code", "-1");
+}
+
+ICMPService::ICMPService(const FWObject *root) : Service(root) 
+{
+    setStr("type", "-1");
+    setStr("code", "-1");
+}
+
 ICMPService::~ICMPService() {}
 
-void ICMPService::fromXML(xmlNodePtr root)
+string ICMPService::getProtocolName()    {    return "icmp";}
+int    ICMPService::getProtocolNumber()  {    return 1;     }
+
+void ICMPService::fromXML(xmlNodePtr root) throw(FWException)
 {
     FWObject::fromXML(root);
 
     const char *n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("type")));
     assert(n!=NULL);
     setStr("type", n);
+    FREEXMLBUFF(n);
 
     n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("code")));
-    if(n!=NULL) setStr("code", n);
+    if(n!=NULL)
+    {
+        setStr("code", n);
+        FREEXMLBUFF(n);
+    }
 }
 
 
