@@ -8,6 +8,7 @@ using namespace libfwbuilder;
 const char *IntervalGroup::TYPENAME={"IntervalGroup"};
 
 IntervalGroup::IntervalGroup() : Group() {}
+IntervalGroup::IntervalGroup(const FWObject *root) : FWObject(root) {}
 
 IntervalGroup::~IntervalGroup() {}
 
@@ -20,8 +21,10 @@ bool  IntervalGroup::validateChild(FWObject *o)
 	     otype=="IntervalRef" ));
 }
 
-FWReference* IntervalGroup::createRef( FWObject* obj )
+FWReference* IntervalGroup::createRef()
 {
-    return new FWIntervalReference(obj);
+    FWIntervalReference *ref=new FWIntervalReference();
+    ref->setPointer(this);
+    return ref;
 }
 
