@@ -2,11 +2,11 @@
 
                           Firewall Builder
 
-                 Copyright (C) 2000 Vadim Kurland
+                 Copyright (C) 2000 NetCitadel, LLC
 
   Author:  Vadim Kurland     vadim@vk.crocodile.org
 
-  $Id: NAT.cc,v 1.3 2001/12/19 12:46:54 lord Exp $
+  $Id: NAT.cc,v 1.8 2002/09/08 19:28:16 vkurland Exp $
 
 
   This program is free software which we release under the GNU General Public
@@ -24,53 +24,36 @@
 
 */
 
+#include <assert.h>
 
 #include <fwbuilder/NAT.hh>
 #include <fwbuilder/Rule.hh>
-
 #include <fwbuilder/FWOptions.hh>
-
 #include <fwbuilder/FWObjectDatabase.hh>
 
 using namespace libfwbuilder;
 
 const char *NAT::TYPENAME={"NAT"};
-
-static const char *rels[]={
-    "OSrc",
-    "ODst",
-    "OSrv",
-    "TSrc",
-    "TDst",
-    "TSrv",
-    "Com",
-    NULL
-};
  
 NAT::NAT() : RuleSet() 
 {
     remStr("name");
     remStr("comment");
-
-    rule_elements=rels;
 }
 
+NAT::NAT(const FWObject *root) : RuleSet(root) 
+{
+    remStr("name");
+    remStr("comment");
+}
 
 NAT::~NAT()  {}
 
 Rule* NAT::createRule()
 {
-    Rule      *r;
-
-    r= new NATRule();
-    r->add( new NATRuleOptions() );
-
-    return r;
+    FWObjectDatabase* db=dynamic_cast<FWObjectDatabase*>(getRoot());
+    assert(db!=NULL);
+    return Rule::cast( db->create(NATRule::TYPENAME, true) );
 }
 
 
-
-
-
-
-
