@@ -2,11 +2,11 @@
 
                           Firewall Builder
 
-                 Copyright (C) 2000 Vadim Kurland
+                 Copyright (C) 2000 NetCitadel, LLC
 
   Author:  Vadim Kurland     vadim@vk.crocodile.org
 
-  $Id: XMLTools.cc,v 1.9 2001/12/28 05:32:31 lord Exp $
+  $Id: XMLTools.cc,v 1.34 2003/12/31 06:51:36 vkurland Exp $
 
 
   This program is free software which we release under the GNU General Public
@@ -31,6 +31,11 @@
 
 #include <string.h>
 #include <unistd.h>
+#include <sys/stat.h>
+#include <fcntl.h>
+#include <unistd.h>
+#include <errno.h>
+#include <assert.h>
 
 #ifdef HAVE_LIBXSLT_XSLTCONFIG_H
 # include <libxslt/xsltconfig.h>
@@ -43,16 +48,23 @@
 
 #include <glib.h>
 
+#include <iostream>
+
 #undef FW_XMLTOOLS_VERBOSE
 //#define FW_XMLTOOLS_VERBOSE
 
-#define DTD_LOAD_BITS (XML_DETECT_IDS|XML_COMPLETE_ATTRS)
+#define DTD_LOAD_BITS (1|XML_DETECT_IDS|XML_COMPLETE_ATTRS)
 
 using namespace std;
 using namespace libfwbuilder;
 
+#ifndef __MINGW32__
 extern int xmlDoValidityCheckingDefaultValue ;
 extern int xmlLoadExtDtdDefaultValue         ;
+#else
+extern __declspec(dllimport) int xmlDoValidityCheckingDefaultValue ;
+extern __declspec(dllimport) int xmlLoadExtDtdDefaultValue         ;
+#endif
 
 /*
  * This mutex protects access to XML parser.
@@ -120,7 +132,7 @@ xmlNodePtr XMLTools::getXmlNodeByPath(xmlNodePtr r, const char *path)
     s1=path_copy;
     if (*s1=='/') {
 	res=getXmlNodeByPath(r,s1+1);
-	delete path_copy;
+	delete[] path_copy;
 	return(res);
     }
 
@@ -135,14 +147,14 @@ xmlNodePtr XMLTools::getXmlNodeByPath(xmlNodePtr r, const char *path)
 		if ( xmlIsBlankNode(cur) ) continue;
 		res=getXmlNodeByPath(cur,cptr);
 		if (res) {
-		    delete path_copy;
+		    delete[] path_copy;
 		    return(res);
 		}
 	    }
 	} else
 	    res=r;
     }
-    delete path_copy;
+    delete[] path_copy;
     return(res);
 }
 
@@ -153,7 +165,8 @@ xmlExternalEntityLoader XMLTools::defaultLoader = NULL;
  * This is global variable used in 'fwbExternalEntityLoader'
  * parser callback. It is protected by 'xml_parser_mutex'.
  */
-static string current_template_dir;
+
+static char* current_template_dir=NULL;
 
 xmlParserInputPtr fwbExternalEntityLoader(const char *URL, 
                                           const char *ID,
@@ -162,7 +175,7 @@ xmlParserInputPtr fwbExternalEntityLoader(const char *URL,
     xmlParserInputPtr ret;
 
 #ifdef FW_XMLTOOLS_VERBOSE
-    cerr << "ENTITY: " << URL << " " << ID << endl;
+    cerr << "ENTITY: " << URL << " " << string((ID)?ID:"(null)") << endl;
 #endif
 
     // Try to load it as file from template directory
@@ -188,22 +201,60 @@ void XMLTools::initXMLTools()
     xml_parser_mutex     = g_mutex_new();
     xslt_processor_mutex = g_mutex_new();
     defaultLoader = xmlGetExternalEntityLoader();
+    current_template_dir=cxx_strdup("");
     xmlSetExternalEntityLoader(fwbExternalEntityLoader);
 }
 
+/*
+ *  I use xmlParseMemory because libxml2 v2.6.2 broke xmlParseFile. If
+ *  file name had a full path in it, it won't load and gave an error
+ *  like this:
+ *
+ *  I/O warning : failed to load external entity "/resources.xml"
+ *
+ *  I do not check if xmlParseMemory succeeded in this method because
+ *  I may need to throw different exceptions in different method using
+ *  loadAndParseFile. loadAndParseFile just returns whatever
+ *  xmlParseMemory returned, this works better in a sense that
+ *  loadAndParseFile should be just a wrapper for xmlParseMemory and
+ *  should replicate its behavior.
+ *
+ */
+xmlDocPtr XMLTools::loadAndParseFile(const std::string &rfile) throw(FWException)
+{
+    struct stat stt;
+    int         fd;
+
+    if (stat( rfile.c_str() , &stt )!=0 || (fd=open(rfile.c_str(),O_RDONLY))<0)
+        throw FWException("Could not read file "+rfile);
+
+    char *buf=(char*)malloc(stt.st_size);
+
+    read(fd,buf,stt.st_size);
+    close(fd);
+
+    xmlDocPtr doc = xmlParseMemory(buf, stt.st_size);
+    memset(buf,0,stt.st_size);
+    free(buf);
+
+    return doc;
+}
+
 xmlDocPtr XMLTools::parseFile(const string &file_name, 
                               bool use_dtd, const string &template_dir) throw(FWException)
 {
     g_mutex_lock(xml_parser_mutex);    
 
-    current_template_dir=template_dir;
+    if (current_template_dir!=NULL) delete[] current_template_dir;
+    current_template_dir=cxx_strdup(template_dir.c_str());
     
     xmlDoValidityCheckingDefaultValue = use_dtd?1:0;
     xmlLoadExtDtdDefaultValue         = use_dtd?DTD_LOAD_BITS:0;
     
     string errors;
     xmlSetGenericErrorFunc (&errors, xslt_error_handler);
-    xmlDocPtr doc = xmlParseFile(file_name.c_str()); 
+//    xmlDocPtr doc = xmlParseFile(file_name.c_str()); 
+    xmlDocPtr doc = loadAndParseFile(file_name);
     xmlSetGenericErrorFunc (NULL, NULL);
 
     g_mutex_unlock(xml_parser_mutex);    
@@ -226,7 +277,12 @@ xmlDocPtr XMLTools::loadFile(const string &data_file ,
                              ) throw(FWException)
 {
 #ifdef FW_XMLTOOLS_VERBOSE
-    cerr << "Loading file: " << data_file << endl;
+    cerr << "Loading file: " << data_file
+         << " type: " << type
+         << " dtd_file: " << dtd_file
+         << " template_dir: " << template_dir
+         << " current_version: " << current_version
+         << endl;
 #endif
     
     if(access(data_file.c_str() , R_OK )!=0)
@@ -242,10 +298,12 @@ xmlDocPtr XMLTools::loadFile(const string &data_file ,
     xmlDocPtr newdoc=convert(doc, data_file, type, template_dir, current_version);
     if(newdoc)
     {
-        const string upgrade_msg="Data file '"+data_file+"' was converted from format used in older versions of the product.\n"
-"Do you want to save file in new format? If you choose YES file on disk will be updated and\n"
-"backup copy in old format will be made in the same directory with .bak extension. If you select\n"
-"NO file loading operation will be cancelled.";
+        const string upgrade_msg="The file '"+data_file+"' was saved with\n\
+an older version of Firewall Builder.  Opening it in this version will\n\
+cause it to be upgraded, which may prevent older versions of the program\n\
+from reading it. Backup copy of your file in the old format will be made\n\
+in the same directory with extension '.bak'. Are you sure you want to open it?";
+
 
         if(!(*upgrade)(upgrade_msg))
         {
@@ -275,7 +333,7 @@ xmlDocPtr XMLTools::loadFile(const string &data_file ,
             if(rename(backup_file.c_str(), data_file.c_str()))
             {
                 throw FWException(ex.toString()+"\nRestoring backup copy failed "+
-                                  "your old data could be found in file: '"+backup_file+"'");
+                                  "your old data could be found in the file: '"+backup_file+"'");
             } else
                 throw;
         }
@@ -294,6 +352,11 @@ void XMLTools::setDTD(xmlDocPtr doc,
                       const string &type_name, 
                       const string &dtd_file) throw(FWException)
 {
+#ifdef FW_XMLTOOLS_VERBOSE
+    cerr << "XMLTools::setDTD: type_name=" << type_name << " dtd_file=" << dtd_file << endl;
+#endif
+  
+
     xmlCreateIntSubset(doc, STRTOXMLCAST(type_name), 
                        NULL, 
                        STRTOXMLCAST(dtd_file)
@@ -304,16 +367,31 @@ void XMLTools::setDTD(xmlDocPtr doc,
 
     xmlDoValidityCheckingDefaultValue = 1;
     xmlLoadExtDtdDefaultValue         = DTD_LOAD_BITS;
+    xmlSubstituteEntitiesDefaultValue = 1;
 
     string errors;
     xmlSetGenericErrorFunc (&errors, xslt_error_handler);
 
     try
     {
-        xmlValidCtxt vctxt;
-        if(xmlValidateDocument(&vctxt, doc)!=1)
-            throw FWException("DTD validation stage 2 failed");
+/*
+ * This broke with libxml 2.6.4. Tests seem to rule out bug inside
+ * libxml2 (used their example program "tree2.c" and added similar
+ * fragment for validation, it worked), so it must be something in our
+ * code. I can't seem to find the problem though. 
+ *
+ * We recreate the tree from the objects in the memory, so doing
+ * validation here is mostly a double check. It should be relatively
+ * safe to just skip validation until I figure out what's wrong with
+ * it.
 
+        xmlValidCtxt vctxt;
+        vctxt.userData = &errors;
+        vctxt.error    = xslt_error_handler;
+        vctxt.warning  = xslt_error_handler;
+        if(xmlValidateDocument(&vctxt, doc)!=1)
+            throw FWException(string("DTD validation stage 2 failed with following errors:\n")+errors);
+*/
         xmlSetGenericErrorFunc (NULL, NULL);
         g_mutex_unlock(xml_parser_mutex);    
 
@@ -337,15 +415,88 @@ void XMLTools::saveFile(xmlDocPtr doc,
     
     setDTD(doc, type_name, dtd_file);
 
-#ifdef HAVE_XMLSAVEFORMATFILE
-    if(xmlSaveFormatFile(file_name.c_str(), doc, 1)==-1)
+#ifdef HAVE_XMLSAVEFORMATFILEENC
+    if(xmlSaveFormatFileEnc(file_name.c_str(), doc, "utf-8", 1)==-1)
 #else
-    if(xmlSaveFile(file_name.c_str(), doc)==-1)
+    if(xmlSaveFileEnc(file_name.c_str(), doc, "utf-8")==-1)
 #endif
         throw FWException("Error saving XML file: "+file_name);
 }
 
 
+void XMLTools::transformFileToFile(const string &src_file,
+				   const string &stylesheet_file,
+				   const char **params,
+				   const string &dst_file
+) throw(FWException)
+{
+    string xslt_errors;
+    xsltStylesheetPtr ss = NULL;
+    xmlDocPtr doc, res;
+
+
+    g_mutex_lock(xslt_processor_mutex);    
+    g_mutex_lock(xml_parser_mutex);    
+    
+    xsltSetGenericErrorFunc(&xslt_errors, xslt_error_handler);
+    xmlSetGenericErrorFunc (&xslt_errors, xslt_error_handler);
+
+    // Following line is workaround for bug #73088 in Gnome
+    // bugzilla. To be removed than it will be fixed.
+    xsltSetGenericDebugFunc (&xslt_errors, xslt_error_handler);
+
+    xmlSubstituteEntitiesDefault(1);
+    xmlLoadExtDtdDefaultValue = DTD_LOAD_BITS;
+    ss = xsltParseStylesheetFile(STRTOXMLCAST(stylesheet_file));
+
+    if(!ss)
+    {
+        xsltSetGenericErrorFunc(NULL, NULL);
+        xmlSetGenericErrorFunc (NULL, NULL);
+
+        // Following line is workaround for bug #73088 in Gnome
+        // bugzilla. To be removed than it will be fixed.
+        xsltSetGenericDebugFunc (NULL, NULL);
+        g_mutex_unlock(xml_parser_mutex);    
+        g_mutex_unlock(xslt_processor_mutex);    
+        throw FWException("File conversion error: Error loading stylesheet: "+stylesheet_file+
+          (xslt_errors.length()?(string("\nXSLT reports: \n")+xslt_errors):string(""))
+        );
+    }
+    
+
+//    doc = xmlParseFile(src_file.c_str());
+    doc = loadAndParseFile(src_file);
+
+    res = xsltApplyStylesheet(ss, doc, params);
+    xsltSaveResultToFilename(dst_file.c_str(), res, ss, 0);
+
+    xsltSetGenericErrorFunc(NULL, NULL);
+    xmlSetGenericErrorFunc (NULL, NULL);
+
+    // Following line is workaround for bug #73088 in Gnome
+    // bugzilla. To be removed than it will be fixed.
+    xsltSetGenericDebugFunc (NULL, NULL);
+
+    g_mutex_unlock(xml_parser_mutex);    
+    g_mutex_unlock(xslt_processor_mutex);    
+    
+    if(!res)
+    {
+        xsltFreeStylesheet(ss);
+        throw FWException("File conversion Error: Error during conversion: "+stylesheet_file+
+          (xslt_errors.length()?(string("XSLT reports: \n")+xslt_errors):string(""))
+        );
+    }
+
+    xsltFreeStylesheet(ss);
+    xmlFreeDoc(res);
+    xmlFreeDoc(doc);
+
+    xsltCleanupGlobals();
+    xmlCleanupParser();
+}
+
 void XMLTools::transformDocumentToFile(xmlDocPtr doc, 
                                        const string &stylesheet_file,
                                        const char **params,
@@ -359,7 +510,17 @@ void XMLTools::transformDocumentToFile(xmlDocPtr doc,
     
     xsltSetGenericErrorFunc(&xslt_errors, xslt_error_handler);
     xmlSetGenericErrorFunc (&xslt_errors, xslt_error_handler);
+
+    // Following line is workaround for bug #73088 in Gnome
+    // bugzilla. To be removed than it will be fixed.
+    xsltSetGenericDebugFunc (&xslt_errors, xslt_error_handler);
+
     
+#ifdef FW_XMLTOOLS_VERBOSE
+    xmlSaveFormatFileEnc(".backup_copy.xml", doc, "utf-8", 1);
+    cerr << "Backup copy of the XML tree is saved to file .backup_copy.xml" << endl;
+#endif
+
     xmlDoValidityCheckingDefaultValue = 0;
     xmlLoadExtDtdDefaultValue         = 0;
     xsltStylesheetPtr ss = xsltParseStylesheetFile(STRTOXMLCAST(stylesheet_file));
@@ -370,6 +531,10 @@ void XMLTools::transformDocumentToFile(xmlDocPtr doc,
     {
         xsltSetGenericErrorFunc(NULL, NULL);
         xmlSetGenericErrorFunc (NULL, NULL);
+
+        // Following line is workaround for bug #73088 in Gnome
+        // bugzilla. To be removed than it will be fixed.
+        xsltSetGenericDebugFunc (NULL, NULL);
         g_mutex_unlock(xml_parser_mutex);    
         g_mutex_unlock(xslt_processor_mutex);    
         throw FWException("File conversion error: Error loading stylesheet: "+stylesheet_file+
@@ -381,6 +546,11 @@ void XMLTools::transformDocumentToFile(xmlDocPtr doc,
 
     xsltSetGenericErrorFunc(NULL, NULL);
     xmlSetGenericErrorFunc (NULL, NULL);
+
+    // Following line is workaround for bug #73088 in Gnome
+    // bugzilla. To be removed than it will be fixed.
+    xsltSetGenericDebugFunc (NULL, NULL);
+
     g_mutex_unlock(xml_parser_mutex);    
     g_mutex_unlock(xslt_processor_mutex);    
     
@@ -413,7 +583,11 @@ xmlDocPtr XMLTools::transformDocument(xmlDocPtr doc,
         
     xsltSetGenericErrorFunc(&xslt_errors, xslt_error_handler);
     xmlSetGenericErrorFunc (&xslt_errors, xslt_error_handler);
-    
+
+    // Following line is workaround for bug #73088 in Gnome
+    // bugzilla. To be removed than it will be fixed.
+    xsltSetGenericDebugFunc (&xslt_errors, xslt_error_handler);
+
     xmlDoValidityCheckingDefaultValue = 0;
     xmlLoadExtDtdDefaultValue         = 0;
     xsltStylesheetPtr ss = xsltParseStylesheetFile(STRTOXMLCAST(stylesheet_file));
@@ -423,6 +597,10 @@ xmlDocPtr XMLTools::transformDocument(xmlDocPtr doc,
     {
         xsltSetGenericErrorFunc(NULL, NULL);
         xmlSetGenericErrorFunc (NULL, NULL);
+        // Following line is workaround for bug #73088 in Gnome
+        // bugzilla. To be removed than it will be fixed.
+        xsltSetGenericDebugFunc (NULL, NULL);
+    
         g_mutex_unlock(xml_parser_mutex);    
         g_mutex_unlock(xslt_processor_mutex);    
         throw FWException("File conversion error: Error loading stylesheet: "+stylesheet_file+
@@ -435,6 +613,10 @@ xmlDocPtr XMLTools::transformDocument(xmlDocPtr doc,
     xsltFreeStylesheet(ss);
     xsltSetGenericErrorFunc(NULL, NULL);
     xmlSetGenericErrorFunc (NULL, NULL);
+    // Following line is workaround for bug #73088 in Gnome
+    // bugzilla. To be removed than it will be fixed.
+    xsltSetGenericDebugFunc (NULL, NULL);
+    
     g_mutex_unlock(xml_parser_mutex);    
     g_mutex_unlock(xslt_processor_mutex);    
     
@@ -462,9 +644,10 @@ xmlDocPtr XMLTools::convert(xmlDocPtr doc,
 	xmlFreeDoc(doc);
         throw FWException("XML file '"+file_name+ "' have invalid structure.");
     }
-    
+
+    string vers;
     const char *v=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("version")));
-    if(!v)
+    if(v==NULL)
     {
         // no version.
         v="0.8.7"; // at this version attribute was introducted
@@ -472,21 +655,30 @@ xmlDocPtr XMLTools::convert(xmlDocPtr doc,
                    TOXMLCAST("version") , 
                    TOXMLCAST(v));
         res=doc; // changed
+        vers=v;
+    } else
+    {
+        vers=v;
+        FREEXMLBUFF(v);
     }
 
+#ifdef FW_XMLTOOLS_VERBOSE
+        cerr << "File reports version : " << vers << endl;
+#endif
+
     int c;
-    while(v && (c=version_compare(current_version,v))!=0)
+    while(!vers.empty() && (c=version_compare(current_version,vers))!=0)
     {
         if(c<0)
             throw FWException(string("Data file '"+file_name+ "' is in format, \nof version newer that current product version."));
             
-        string oldversion=v;
+        string oldversion=vers;
         
 #ifdef FW_XMLTOOLS_VERBOSE
         cerr << "Converting from version: " << oldversion << endl;
 #endif
 
-        string fname=template_dir+"/migration"+"/"+v+"/"+type_name+".xslt";
+        string fname=template_dir+"/migration"+"/"+vers+"/"+type_name+".xslt";
 
         if(access(fname.c_str() , R_OK )!=0) 
         {
@@ -499,6 +691,7 @@ xmlDocPtr XMLTools::convert(xmlDocPtr doc,
             res=transformDocument(doc, fname, NULL);
         } catch(FWException &ex)
         {
+            ex.getProperties()["failed_transformation"]=fname;
             xmlFreeDoc(doc);
             throw;
         }
@@ -513,16 +706,19 @@ xmlDocPtr XMLTools::convert(xmlDocPtr doc,
         }
 
         v=FROMXMLCAST(xmlGetProp(root, TOXMLCAST("version")));
-        if(!v)
+        if(v==NULL)
         {
             xmlFreeDoc(doc);
             throw FWException("File '"+file_name+ "' conversion error: converted to unknown version.");
-        }
+        } 
+
+        vers=v;
+        FREEXMLBUFF(v);
         
-        if(version_compare(v, oldversion) <= 0)
+        if(version_compare(vers, oldversion) <= 0)
         {
             xmlFreeDoc(doc);
-            throw FWException("File '"+file_name+ "' conversion error: conversion did not advanced version number!.");
+            throw FWException("File '"+file_name+ "' conversion error: conversion did not advance version number!.");
         }
     }
 
@@ -590,6 +786,39 @@ string XMLTools::unquote_linefeeds(const string &s)
     return res;
 }
 
+string XMLTools::cleanForUTF8(const string &str)
+{
+    string res;
+    
+    if ( xmlCheckUTF8(TOXMLCAST(str.c_str()))==0 )
+    {
+        for (string::size_type i=0; i<str.size(); ++i)
+        {
+            char c[2];
+            c[1]='\0';
+            c[0]=str[i];
+            if ( xmlCheckUTF8(TOXMLCAST(c))==0 )
+            {
+                res += "?";
+            } else  res += c;
+        }
+
+    } else res=str;
+
+    return res;
+}
+
+string XMLTools::cleanForNVTASCII(const string &str)
+{
+    string res=str;
+    for (string::size_type i=0; i<res.size(); ++i)
+    {
+        unsigned int u=res[i];
+        if (u>127) res[i]='?';
+    }
+    return res;
+}
+
 #undef DTD_LOAD_BITS
 
 #undef FW_XMLTOOLS_VERBOSE
