@@ -2,11 +2,11 @@
 
                           Firewall Builder
 
-                 Copyright (C) 2001 Vadim Zaliva, Vadim Kurland
+                 Copyright (C) 2001 NetCitadel, LLC
 
   Author:  Vadim Zaliva lord@crocodile.org
 
-  $Id: dns.cc,v 1.3 2001/12/19 12:47:20 lord Exp $
+  $Id: dns.cc,v 1.42 2003/06/21 07:29:15 vkurland Exp $
 
 
   This program is free software which we release under the GNU General Public
@@ -24,9 +24,17 @@
 
 */
 
+#include <fwbuilder/libfwbuilder-config.h>
+
 #include <fwbuilder/dns.hh>
 
-#include <netdb.h>
+#ifndef __MINGW32__
+#ifdef HAVE_LWRES
+  #include <lwres/lwres.h>
+  #include <lwres/netdb.h>
+#else
+  #include <netdb.h>
+#endif
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
@@ -37,9 +45,14 @@
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <arpa/nameser.h>
+#include <pthread.h>
+#else
+#include <winsock2.h>
+#endif //__MINGW32__
 #include <unistd.h>
 
 #include <memory>
+#include <sstream>
 
 #include <fwbuilder/ThreadTools.hh>
 
@@ -48,12 +61,47 @@ using namespace libfwbuilder;
 
 const size_t DNS::RSP_BUF_SIZE=64*1024;
 
+Mutex *DNS::gethostbyname_mutex;
+Mutex *DNS::gethostbyaddr_mutex;
+
+void libfwbuilder::init_dns() throw(FWException)
+{
+    DNS::init();
+
+#ifdef HAVE_LWRES_GETIPNODE
+    struct hostent *hp;
+    int    herr;
+    int    ret;
+    char   addrbuf[16];
+
+    ret = inet_pton(AF_INET, "127.0.0.1", addrbuf);
+
+    hp = lwres_getipnodebyaddr(addrbuf, 4, AF_INET, &herr);
+    if (hp==NULL && herr==NO_RECOVERY)
+        throw FWException(
+"Lightweight resolver library test failed; perhaps lwresd is not running.\nDNS queries won't work.");
+
+#endif
+
+}
+
+void DNS::init()
+{
+#if !defined( HAVE_FUNC_GETHOSTBYNAME_R_6) && !defined(HAVE_FUNC_GETHOSTBYNAME_R_5)
+    gethostbyname_mutex = new Mutex();
+    gethostbyaddr_mutex = new Mutex();
+#endif 
+}
+
 DNS::DNS()
 {
 }
 
 string DNS::getErrorMessage(int rcode)
 {
+#if !defined(HAVE_GOODLIBRESOLV)
+    throw FWException("This feature is not implemented on your OS.");
+#else
     map<int, string> error_messages;
     
     // Define some common error codes
@@ -79,10 +127,15 @@ string DNS::getErrorMessage(int rcode)
         sprintf(buf,"DNS Error '%d'", rcode);
         return buf;
     }
+#endif
 }
 
 HostEnt DNS::getHostByAddr(const IPAddress &addr, int retries_, int timeout_) throw(FWException)
 {
+#if !defined(HAVE_GOODLIBRESOLV)
+# warning "getHostByAddr will use default system timeout because bind library was not found on the system "
+    return getHostByAddr(addr);
+#else
     struct __res_state res;
 
     if(res_ninit(&res)==-1)
@@ -147,6 +200,7 @@ HostEnt DNS::getHostByAddr(const IPAddress &addr, int retries_, int timeout_) th
         } 
     }
     return v;
+#endif
 }
 
 
@@ -160,14 +214,22 @@ HostEnt DNS::getHostByAddr(const IPAddress &addr) throw(FWException)
     char *tmphstbuf = (char *)malloc(hstbuflen);
 
     struct in_addr naddr;
-    naddr.s_addr=addr;
+    naddr.s_addr=addr.to32BitInt();
 
+#ifdef HAVE_LWRES_GETIPNODE
+    struct hostent *res;
+    char   addrbuf[16];
+    int    ret=inet_pton(AF_INET, addr.toString().c_str(), addrbuf);
+    hp = res = lwres_getipnodebyaddr(addrbuf, 4, AF_INET, &herr);
+    if (res!=NULL) herr=0;
+
+#elif HAVE_GETHOSTBYADDR_R_8
     int res;
-#ifdef DEBIAN
+# ifdef GETHOSTBYADDR_FIRST_ARG_CHARPTR
     while((res = gethostbyaddr_r((const char *)&naddr, sizeof(naddr),
-#else
-     while((res = gethostbyaddr_r(&naddr, sizeof(naddr),
-#endif
+# else
+    while((res = gethostbyaddr_r(&naddr, sizeof(naddr),
+# endif
                                  AF_INET,
                                  &hostbuf,
                                  tmphstbuf, hstbuflen,
@@ -178,9 +240,39 @@ HostEnt DNS::getHostByAddr(const IPAddress &addr) throw(FWException)
         hstbuflen *= 2;
         tmphstbuf = (char *)realloc(tmphstbuf, hstbuflen);
     }
-    if(res || herr)
+#elif HAVE_GETHOSTBYADDR_R_7
+    struct hostent *res;
+    while((hp = res = gethostbyaddr_r((const char *)&naddr, sizeof(naddr),
+                                 AF_INET,
+                                 &hostbuf,
+                                 tmphstbuf, hstbuflen,
+                                 &herr)
+          ) && (herr == ERANGE))
     {
+        hstbuflen *= 2;
+        tmphstbuf = (char *)realloc(tmphstbuf, hstbuflen);
+    }
+#else
+# warning "No gethostbyaddr_r(); all DNS calls will be serialized (slow!)"
+
+    struct hostent *res;
+
+    gethostbyaddr_mutex->lock();
+    hp = res = gethostbyaddr((const char *)&naddr, 
+                             sizeof(naddr), 
+                             AF_INET
+    );
+    herr = 0;
+
+#endif
+
+    if(!res || herr)
+    {
+#if !defined(HAVE_LWRES_GETIPNODE) && !defined(HAVE_GETHOSTBYADDR_R_8) && !defined(HAVE_GETHOSTBYADDR_R_7)
+        gethostbyaddr_mutex->unlock();
+#endif
         free(tmphstbuf);
+
         throw FWException(string("Hostname of address: '")+IPAddress(&naddr).toString()+"' not found");
     } 
     
@@ -191,19 +283,32 @@ HostEnt DNS::getHostByAddr(const IPAddress &addr) throw(FWException)
             v.aliases.insert(string(*p));
 
     free(tmphstbuf);
+#if !defined(HAVE_LWRES_GETIPNODE) && !defined(HAVE_GETHOSTBYADDR_R_8) && !defined(HAVE_GETHOSTBYADDR_R_7)
+        gethostbyaddr_mutex->unlock();
+#endif
     return v;
 }
 
 vector<IPAddress> DNS::getHostByName(const string &name) throw(FWException)
 {
     struct hostent hostbuf;
-    struct hostent *hp;
-    int herr;
-    
-    size_t hstbuflen = 1024; 
-    char *tmphstbuf = (char *)malloc(hstbuflen);
+    struct hostent *hp=NULL;
+    char  *tmphstbuf=NULL;
+    size_t hstbuflen = 1024;
+    int    herr;
 
-#ifdef HAVE_FUNC_GETHOSTBYNAME_R_6
+#ifdef HAVE_LWRES_GETIPNODE
+
+    hp=lwres_getipnodebyname(name.c_str(), AF_INET, AI_V4MAPPED + AF_INET6, &herr);
+    if(!hp)
+    {
+        throw FWException("Host or network '"+name+"' not found.\nResolver returned error \""+
+                          lwres_hstrerror(herr)+"\"");
+    }
+
+#elif HAVE_FUNC_GETHOSTBYNAME_R_6
+
+    tmphstbuf = (char *)malloc(hstbuflen);
     int res;
     while((res = gethostbyname_r(name.c_str(), &hostbuf,tmphstbuf,hstbuflen,&hp,&herr))
           && (herr == ERANGE))
@@ -211,27 +316,39 @@ vector<IPAddress> DNS::getHostByName(const string &name) throw(FWException)
         hstbuflen *= 2;
         tmphstbuf = (char *)realloc(tmphstbuf, hstbuflen);
     }
-    if(res)
+    if(res || !hp)
     {
         free(tmphstbuf);
         throw FWException("Host or network '"+name+"' not found");
     } 
-# else
-# ifdef HAVE_FUNC_GETHOSTBYNAME_R_5
+#elif HAVE_FUNC_GETHOSTBYNAME_R_5
+
+    tmphstbuf = (char *)malloc(hstbuflen);
     while(!(hp = gethostbyname_r(name.c_str(), &hostbuf, tmphstbuf, hstbuflen, &herr))
           && (herr == ERANGE))
     {
         hstbuflen *= 2;
         tmphstbuf = (char *)realloc(tmphstbuf, hstbuflen);
     }
+
     if(!hp)
     {
         free(tmphstbuf);
         throw FWException("Host or network '"+name+"' not found");
     } 
-# else
-#   error "unsupported gethostbyname_r()"
-# endif
+#else
+# warning "No gethostbyname_r() all DNS calls will be serialized (slow!)"
+    
+    tmphstbuf = NULL;
+    
+    gethostbyname_mutex->lock();
+    hp = gethostbyname(name.c_str());
+    if(!hp)
+    {
+        gethostbyname_mutex->unlock();
+        throw FWException("Host or network '"+name+"' not found");
+    }
+
 #endif
     
     vector<IPAddress> v;
@@ -241,16 +358,35 @@ vector<IPAddress> DNS::getHostByName(const string &name) throw(FWException)
             v.push_back(IPAddress((struct in_addr *)(*p)));
     } catch(const FWException &e)
     {
-        free(tmphstbuf);
+        if(tmphstbuf)
+            free(tmphstbuf);
+
+#ifdef HAVE_LWRES_GETIPNODE
+        lwres_freehostent(hp);
+#endif
+#if !defined( HAVE_LWRES_GETIPNODE ) && !defined( HAVE_FUNC_GETHOSTBYNAME_R_6) && !defined(HAVE_FUNC_GETHOSTBYNAME_R_5)
+        gethostbyname_mutex->unlock();
+#endif
         throw;
     }
-    free(tmphstbuf);
+    if(tmphstbuf)
+       free(tmphstbuf);
+
+#ifdef HAVE_LWRES_GETIPNODE
+        lwres_freehostent(hp);
+#endif
+#if !defined( HAVE_LWRES_GETIPNODE ) && !defined( HAVE_FUNC_GETHOSTBYNAME_R_6) && !defined(HAVE_FUNC_GETHOSTBYNAME_R_5)
+    gethostbyname_mutex->unlock();
+#endif
+
     return v;
 }
 
-
-multimap<string, IPAddress> DNS::getNS(const string &domain, Logger *logger, int retries_, int timeout_) throw(FWException)
+multimap<string, IPAddress> DNS::getNS(const string &domain, Logger *logger,SyncFlag *stop_program, int retries_, int timeout_) throw(FWException)
 {
+#if !defined(HAVE_GOODLIBRESOLV)
+    throw FWException("This feature is not implemented on your OS.");
+#else
     struct __res_state res;
     
     if(res_ninit(&res)==-1)
@@ -259,7 +395,7 @@ multimap<string, IPAddress> DNS::getNS(const string &domain, Logger *logger, int
     res.retrans = timeout_;
     res.retry   = retries_;
 
-    check_stop();
+    CHECK_STOP
 
     auto_ptr<unsigned char> answer(new unsigned char[RSP_BUF_SIZE]);
     
@@ -271,7 +407,7 @@ multimap<string, IPAddress> DNS::getNS(const string &domain, Logger *logger, int
                           answer.get(), 
                           RSP_BUF_SIZE);
 
-    check_stop();
+    CHECK_STOP
     
     if(len<0)
         throw FWException("Error returned while quering domain NS records");
@@ -284,19 +420,19 @@ multimap<string, IPAddress> DNS::getNS(const string &domain, Logger *logger, int
     if(ns_initparse(answer.get(), len, &handle) < 0) 
         throw FWException("Zone parse error in initparse");
     
-    check_stop();
+    CHECK_STOP
 
     if(ns_msg_getflag(handle, ns_f_rcode) != ns_r_noerror)
         throw FWException(getErrorMessage(ns_msg_getflag(handle, ns_f_rcode)));
     
-    check_stop();
+    CHECK_STOP
 
     if(ns_msg_count(handle, ns_s_an) == 0)
         throw FWException("Answer contains to records");
     
     while(true) 
     {        
-        check_stop();
+        CHECK_STOP
         ns_rr rr;
         if(ns_parserr(&handle, ns_s_an, -1, &rr))
         {
@@ -306,15 +442,15 @@ multimap<string, IPAddress> DNS::getNS(const string &domain, Logger *logger, int
                 break;
         }
         
-        check_stop();
+        CHECK_STOP
         if(ns_rr_type(rr)==ns_t_ns && ns_rr_class(rr)==ns_c_in)
         {
             char dn[NS_MAXDNAME];
             if(dn_expand(answer.get(), answer.get() + len, ns_rr_rdata(rr), dn, sizeof(dn))<0)
                 throw FWException("A record parse error in parserr");
-            check_stop();
+            CHECK_STOP
             vector<IPAddress> a=DNS::getHostByName(dn);
-            check_stop();
+            CHECK_STOP
             for(vector<IPAddress>::iterator i=a.begin();i!=a.end();++i)
                 v.insert(pair<string, IPAddress>(string(dn), (*i)));
         } 
@@ -322,6 +458,7 @@ multimap<string, IPAddress> DNS::getNS(const string &domain, Logger *logger, int
     
     *logger << "Succesfuly found " << (int)v.size() << " name servers."  << '\n';
     return v;
+#endif
 }
 
 /**
@@ -329,25 +466,28 @@ multimap<string, IPAddress> DNS::getNS(const string &domain, Logger *logger, int
  * TCP connection to transfer zone established and attempted 
  * only once.
  */
-map<string, set<IPAddress> > DNS::findA(const string &domain, Logger *logger, int retries_, int timeout_) throw(FWException)
+map<string, set<IPAddress> > DNS::findA(const string &domain, Logger *logger,SyncFlag *stop_program, int retries_, int timeout_) throw(FWException)
 {
+#if !defined(HAVE_GOODLIBRESOLV)
+    throw FWException("This feature is not implemented on your OS.");
+#else
     TimeoutCounter timeout(timeout_, "Getting A records");
 
     *logger << "Looking for authoritative servers" << '\n';
     
-    multimap<string, IPAddress> ns=DNS::getNS(domain, logger);
+    multimap<string, IPAddress> ns=DNS::getNS(domain, logger,stop_program);
     if(!ns.size())
         throw FWException("No NS records found");
 
-    check_stop();
+    CHECK_STOP
     timeout.check();
 
-    FWException *last_err;
+    FWException *last_err=NULL; // to avoid compiler warning
     for(multimap<string, IPAddress>::iterator nsi=ns.begin(); nsi!=ns.end(); ++nsi)
     {
         try
         {
-            return findA(domain, (*nsi).second, logger, retries_, timeout.timeLeft());
+            return findA(domain, (*nsi).second, logger,stop_program, retries_, timeout.timeLeft());
             timeout.check();
         } catch(FWException &ex)
         {
@@ -365,6 +505,7 @@ map<string, set<IPAddress> > DNS::findA(const string &domain, Logger *logger, in
     // with no luck.
     // Throw last exception.
     throw *last_err;
+#endif
 }
 
 /**
@@ -372,8 +513,11 @@ map<string, set<IPAddress> > DNS::findA(const string &domain, Logger *logger, in
  * TCP connection to transfer zone established and attempted 
  * only once.
  */
-map<string, set<IPAddress> > DNS::findA(const string &domain, const IPAddress &ns, Logger *logger, int retries_, int timeout_) throw(FWException)
+map<string, set<IPAddress> > DNS::findA(const string &domain, const IPAddress &ns, Logger *logger,SyncFlag *stop_program, int retries_, int timeout_) throw(FWException)
 {
+#if !defined(HAVE_GOODLIBRESOLV)
+    throw FWException("This feature is not implemented on your OS.");
+#else
     TimeoutCounter timeout(timeout_, "Getting A records");
 
     *logger << "Querying server: " << ns.toString() << '\n';
@@ -395,7 +539,7 @@ map<string, set<IPAddress> > DNS::findA(const string &domain, const IPAddress &n
     res.nsaddr_list[0].sin_port   = htons(NAMESERVER_PORT);
     res.nsaddr_list[0].sin_addr   = nsaddress;
 
-    check_stop();
+    CHECK_STOP
     timeout.check();
 
     // Create a query packet for the requested zone name.
@@ -405,7 +549,7 @@ map<string, set<IPAddress> > DNS::findA(const string &domain, const IPAddress &n
                               ns_c_in, ns_t_axfr, NULL,
                               0, 0, buf, sizeof buf);
     
-    check_stop();
+    CHECK_STOP
     timeout.check();
 
     if(msglen<0)
@@ -420,7 +564,7 @@ map<string, set<IPAddress> > DNS::findA(const string &domain, const IPAddress &n
     if (connect(sockFD, (struct sockaddr *)sin, sizeof *sin) < 0) 
         throw FWException("Error connecting to DNS server");
     
-    check_stop();
+    CHECK_STOP
     timeout.check();
 
     // Send length & message for zone transfer
@@ -432,12 +576,12 @@ map<string, set<IPAddress> > DNS::findA(const string &domain, const IPAddress &n
             
     try
     {
-        check_stop();
+        CHECK_STOP
 	timeout.check();
 
         if(write(sockFD, tmp, NS_INT16SZ) != NS_INT16SZ)
             throw FWException("Error 01 sending AXFR query to DNS server");
-        check_stop();
+        CHECK_STOP
 	timeout.check();
 
         *logger << "Sending Query" << '\n';
@@ -451,7 +595,7 @@ map<string, set<IPAddress> > DNS::findA(const string &domain, const IPAddress &n
         *logger << "Reading response." << '\n';
         while(!done)
         {
-            check_stop();
+            CHECK_STOP
 	    timeout.check();
 
             // Read the length of the response.
@@ -459,7 +603,7 @@ map<string, set<IPAddress> > DNS::findA(const string &domain, const IPAddress &n
             size_t amtToRead = INT16SZ;
             while(amtToRead > 0)
             {
-                check_stop();
+                CHECK_STOP
 		timeout.check();
 
 		int numRead=timeout.read(sockFD, cp, amtToRead);
@@ -476,7 +620,7 @@ map<string, set<IPAddress> > DNS::findA(const string &domain, const IPAddress &n
             else if(len<0)
                 throw FWException("Ivalid packet size returned");
 
-            check_stop();
+            CHECK_STOP
 	    timeout.check();
 
             try
@@ -492,7 +636,7 @@ map<string, set<IPAddress> > DNS::findA(const string &domain, const IPAddress &n
             cp = answer.get();
             while(amtToRead > 0)
             {
-                check_stop();
+                CHECK_STOP
 		timeout.check();
 
 		int numRead=timeout.read(sockFD, cp, amtToRead);
@@ -508,7 +652,7 @@ map<string, set<IPAddress> > DNS::findA(const string &domain, const IPAddress &n
 
             ns_msg handle;
                     
-            check_stop();
+            CHECK_STOP
 	    timeout.check();
 
             if(ns_initparse(answer.get(), len, &handle) < 0) 
@@ -522,7 +666,7 @@ map<string, set<IPAddress> > DNS::findA(const string &domain, const IPAddress &n
                     
             while(true) 
             {
-                check_stop();
+                CHECK_STOP
 		timeout.check();
 
                 ns_rr rr;
@@ -534,7 +678,7 @@ map<string, set<IPAddress> > DNS::findA(const string &domain, const IPAddress &n
                         break;
                 }
                         
-                check_stop();
+                CHECK_STOP
 		timeout.check();
 
                 if(ns_rr_type(rr)==ns_t_a && ns_rr_class(rr)==ns_c_in)
@@ -573,6 +717,7 @@ map<string, set<IPAddress> > DNS::findA(const string &domain, const IPAddress &n
     }
     *logger << "Succesfuly found " << (int)v.size() << " hosts."  << '\n';
     return v;
+#endif
 }
 
 DNS_getNS_query::DNS_getNS_query(const string &domain_, int retries_, int timeout_)
@@ -582,9 +727,9 @@ DNS_getNS_query::DNS_getNS_query(const string &domain_, int retries_, int timeou
     timeout = timeout_ ;
 }
 
-void DNS_getNS_query::run_impl(Logger *logger) throw(FWException)
+void DNS_getNS_query::run_impl(Logger *logger,SyncFlag *stop_program) throw(FWException)
 {
-    result = DNS::getNS(domain, logger, retries, timeout);
+    result = DNS::getNS(domain, logger,stop_program, retries, timeout);
 }
 
 DNS_findA_query::DNS_findA_query()
@@ -607,9 +752,9 @@ void DNS_findA_query::init(const string &domain_, const IPAddress &ns_, int retr
         ns      = ns_      ;
 }
 
-void DNS_findA_query::run_impl(Logger *logger) throw(FWException)
+void DNS_findA_query::run_impl(Logger *logger,SyncFlag *stop_program) throw(FWException)
 {
-    result = findA(domain, ns, logger, retries, timeout);
+    result = findA(domain, ns, logger,stop_program, retries, timeout);
 }
 
 DNS_bulkBackResolve_query::DNS_bulkBackResolve_query(set<IPAddress> _ips, 
@@ -624,10 +769,12 @@ DNS_bulkBackResolve_query::DNS_bulkBackResolve_query(set<IPAddress> _ips,
     timeout  = _timeout  ;
     nthreads = _nthreads ;
 
+    pthread_attr_init(&tattr);
 }
 
 DNS_bulkBackResolve_query::~DNS_bulkBackResolve_query()
 {
+    pthread_attr_destroy(&tattr);
 }
 
 void* libfwbuilder::DNS_bulkBackResolve_Thread(void *args)
@@ -635,70 +782,115 @@ void* libfwbuilder::DNS_bulkBackResolve_Thread(void *args)
     void **void_pair=(void**)args;
     DNS_bulkBackResolve_query *p      = static_cast<DNS_bulkBackResolve_query*>(void_pair[0]);
     Logger                    *logger = static_cast<Logger *>(void_pair[1]);
-
-    while(!p->get_stop_program_flag())
+    SyncFlag            *stop_program = static_cast<SyncFlag*>(void_pair[2]);
+    try 
     {
-        p->queue_mutex.lock();
-        if(p->ips.empty())
+        while( true )
         {
-            p->queue_mutex.unlock();
-            break;
-        }
-        IPAddress j=p->ips.front(); p->ips.pop();
-        p->queue_mutex.unlock();
-        
-        try
-        {
-            if(p->get_stop_program_flag())
+            CHECK_STOP
+
+            p->queue_mutex.lock();
+            if(p->ips.empty())
+            {
+                p->queue_mutex.unlock();
                 break;
-            HostEnt he=DNS::getHostByAddr(j, p->retries, p->timeout);
-	    *logger << start << "Resolved  " << j.toString() << ": " << he.name << "\n" << end ;
-            p->result_mutex.lock();
-            p->result[j]=he;
-            p->result_mutex.unlock();
-        } catch(FWException &ex)
-        {
-            *logger << start << "Could not resolve address " << j.toString() << " to a host name, will use generic name\n" << end ;
-            p->failed_mutex.lock();
-            p->failed.insert(j);
-            p->failed_mutex.unlock();
+            }
+            IPAddress j=p->ips.front(); p->ips.pop();
+
+	    *logger << "Trying " << j.toString() 
+		    << ";   " << int(p->ips.size()) << " more in the queue\n";
+
+            p->queue_mutex.unlock();
+        
+            CHECK_STOP
+
+            try
+            {
+
+                HostEnt he=DNS::getHostByAddr(j, p->retries, p->timeout);
+                std::ostringstream str;
+                str << "Resolved  " << j.toString() << ": " << he.name << "\n";
+                *logger << str.str();
+                p->result_mutex.lock();
+                p->result[j]=he;
+                p->result_mutex.unlock();
+            } catch(FWException &ex)
+            {
+                std::ostringstream str;
+                str << "Could not resolve address " << j.toString() 
+		    << " to a host name, will use generic name\n";
+                *logger << str.str();
+                p->failed_mutex.lock();
+                p->failed.insert(j);
+                p->failed_mutex.unlock();
+            }
         }
+    } 
+    catch (FWException &ex)
+    {
+        // do nothing, just break the "while" loop
     }
-    
+
+    *logger << " Worker thread exiting\n";
+
     p->running_mutex.lock   ();
+    p->running_count--;
     p->running_cond.signal  ();
     p->running_mutex.unlock ();
-    
+
     return NULL;
 }
 
-void DNS_bulkBackResolve_query::run_impl(Logger *logger) throw(FWException)
+void DNS_bulkBackResolve_query::run_impl(Logger *logger,SyncFlag *stop_program) throw(FWException)
 {
     queue_mutex.lock();
+
+    running_count = 0;
     for(unsigned int i=0;i<nthreads;i++)
     {
-        void **void_pair = new void*[2];
+        void **void_pair = new void*[3];
         void_pair[0]     = this;
         void_pair[1]     = logger;
+        void_pair[2]     = stop_program;
+
+	*logger << "  Creating worker thread " << "\n";
 
         pthread_t tid;
+        pthread_attr_setdetachstate(&tattr, PTHREAD_CREATE_DETACHED);
+        running_count++;
         pthread_create(&tid, NULL, DNS_bulkBackResolve_Thread, void_pair);
-        pthread_detach(tid);
     }
-    runnning_count = nthreads;
 
     running_mutex.lock();
 
     queue_mutex.unlock();
 
-    while(runnning_count)
+/*
+ *  thread itself decreases running_count just before it calls signal
+ *  on running_cond and thus wakes up main process in this loop. This
+ *  is done this way because of the situation when we have created
+ *  more threads than we have IPs to be resolved. In this case some
+ *  threads would have finished and exited _before_ this loop even
+ *  started. Threads should decrease the counter themselves so that
+ *  loop would wait only for those threads that worked when the loop
+ *  started.
+ *
+ *  Note that mutex is unlocked before falling asleep in wait and then
+ *  locked again when thread is woken up. This ensures that we always
+ *  check running_count when mutex is locked. Threads use the same
+ *  mutex when modify running_count.
+ */
+    while(running_count)
     {
         running_cond.wait(running_mutex);
-        runnning_count--;
     }
+
     running_mutex.unlock();
 
-    check_stop();
+    CHECK_STOP
+
     if(!failed.empty())
         throw FWException("Some of IPs did not resolve");
 }
+
+
