@@ -2,11 +2,11 @@
 
                           Firewall Builder
 
-                 Copyright (C) 2001 Vadim Zaliva, Vadim Kurland
+                 Copyright (C) 2001 NetCitadel, LLC
 
   Author:  Vadim Zaliva lord@crocodile.org
 
-  $Id: snmp.cc,v 1.6 2001/12/28 19:03:31 lord Exp $
+  $Id: snmp.cc,v 1.26 2003/11/10 07:49:54 vkurland Exp $
 
   This program is free software which we release under the GNU General Public
   License. You may redistribute and/or modify this program under the terms
@@ -29,13 +29,17 @@
 
 #include <stdlib.h>
 #include <stdio.h>
-#include <pthread.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <time.h>
 
+#include <fwbuilder/physAddress.hh>
+#include <fwbuilder/IPv4.hh>
 #include <fwbuilder/snmp.hh>
+
+#ifdef UCD_SNMP
 #include <ucd-snmp/parse.h>
+#endif
 
 #include <typeinfo>
 #include <algorithm>
@@ -61,6 +65,12 @@ const char *SNMPQuery::SNMP_ADDR_TABLE       = "ip.ipAddrTable.ipAddrEntry.ipAdE
 const char *SNMPQuery::SNMP_BCAST_TABLE      = "ip.ipAddrTable.ipAddrEntry.ipAdEntBcastAddr" ;
 const char *SNMPQuery::SNMP_AT_TABLE_NET     = "at.atTable.atEntry.atNetAddress";
 const char *SNMPQuery::SNMP_AT_TABLE_PHYS    = "at.atTable.atEntry.atPhysAddress";
+const char *SNMPQuery::IP_MIB_PREFIX         = "IP-MIB::"             ;
+const char *SNMPQuery::IP_MIB_ENTIFINDEX_PREFIX = "IP-MIB::ipAdEntIfIndex";
+
+const char *SNMPQuery::RFC1213_MIB_PREFIX    = "RFC1213-MIB::"        ;
+const char *SNMPQuery::RFC1213_MIB_PREFIX_IPROUTEDEST_PREFIX = 
+						"RFC1213-MIB::ipRouteDest" ;
 
 const char *SNMPQuery::SNMP_SYSNAME          = "system.sysName.0"     ;
 const char *SNMPQuery::SNMP_SYSDESCR         = "system.sysDescr.0"    ;
@@ -123,7 +133,7 @@ SNMPQuery::~SNMPQuery()
 {
 }
 
-void SNMPQuery::fetchAll(Logger *logger) throw(FWException)
+void SNMPQuery::fetchAll(Logger *logger,SyncFlag *stop_program) throw(FWException)
 {
     if(community.empty())
         throw FWException("No SNMP community specified");
@@ -135,24 +145,24 @@ void SNMPQuery::fetchAll(Logger *logger) throw(FWException)
     *logger << "Connecting to " << hostname << "\n";
     c.connect(retries, timeout);
 
-    check_stop();
+    CHECK_STOP
     
-    fetchSysInfo   (logger, &c);
-    check_stop();
+    fetchSysInfo   (logger,stop_program, &c);
+    CHECK_STOP
     
-    fetchInterfaces(logger, &c);
-    check_stop();
+    fetchInterfaces(logger,stop_program, &c);
+    CHECK_STOP
     
-    fetchArpTable  (logger, &c);
-    check_stop();
+    fetchArpTable  (logger,stop_program, &c);
+    CHECK_STOP
 
-    fetchRoutingTable  (logger, &c);
-    check_stop();
+    fetchRoutingTable  (logger,stop_program, &c);
+    CHECK_STOP
 }
 
-void SNMPQuery::fetchArpTable(Logger *logger, SNMPConnection *connection) throw(FWException)
+void SNMPQuery::fetchArpTable(Logger *logger,SyncFlag *stop_program, SNMPConnection *connection) throw(FWException)
 {
-    *logger << "Fetching ARP table\n"; 
+    *logger << "ARP table\n"; 
     arptable.clear();
     
     SNMPConnection *c;
@@ -167,7 +177,7 @@ void SNMPQuery::fetchArpTable(Logger *logger, SNMPConnection *connection) throw(
         if(hostname.empty())
             throw FWException("No SNMP hostname specified");
        
-        check_stop();
+        CHECK_STOP
         c=new SNMPConnection(hostname, community);
         *logger << "Connecting to " << hostname << "\n";
         try
@@ -177,41 +187,50 @@ void SNMPQuery::fetchArpTable(Logger *logger, SNMPConnection *connection) throw(
         {
             // If program was stopped, show this, rather
             // than error.
-            check_stop();
+            CHECK_STOP
             throw;
         }
-        check_stop();
+        CHECK_STOP
     }
+    *logger << "Connected... " << "\n";
 
     vector<SNMPVariable*> v;
     try
     {
+        *logger << "Pulling atNetAddress table... " << "\n";
         multimap<string, SNMPVariable* > nw=c->walk(SNMP_AT_TABLE_NET  );
+        *logger << "Pulling atPhysAddress table... " << "\n";
         multimap<string, SNMPVariable* > pw=c->walk(SNMP_AT_TABLE_PHYS );
+
         try
         {
-            check_stop();
-            int preflen = strlen(SNMP_AT_TABLE_NET);
+            CHECK_STOP
+//            int preflen = strlen(SNMP_AT_TABLE_NET);
             for(multimap<string, SNMPVariable* >::iterator j = nw.begin(); j!=nw.end(); ++j)
             {
-                check_stop();
+                CHECK_STOP
                 if((*j).second->type!=SNMPVariable::snmp_ipaddr)
                 {
                     *logger << "unexpected result type in '" << 
-                    SNMP_AT_TABLE_NET << "'table. Skipping it.\n";
+                    SNMP_AT_TABLE_NET << "' table. Skipping it.\n";
                     continue;
                 }
                 IPAddress ip = dynamic_cast<SNMPVariable_IPaddr*>((*j).second)->getIPAddressValue();
-                string suffix = ((*j).first).substr(preflen);
-                string hwaddr="?";
-                
-                vector<SNMPVariable*> v = c->get(string(SNMP_AT_TABLE_PHYS)+suffix);
-                if(v.size()==1)
-                    if(v[0]->type==SNMPVariable::snmp_string)
-                        hwaddr=dynamic_cast<SNMPVariable_String*>(v[0])->toHexString();
-                
-                SNMPVariable::freeVarList(v);
-                arptable.insert(make_pair(ip,hwaddr));
+
+                *logger << "Learned: " << ip.toString();
+
+                for(multimap<string, SNMPVariable* >::iterator k = pw.begin(); k!=pw.end(); ++k)
+                {
+                    string pa_oid= k->first;
+                    if ( pa_oid.find( ip.toString() )!=string::npos )
+                    {
+                        string hwaddr=dynamic_cast<SNMPVariable_String*>(k->second)->toHexString();
+                        arptable.insert(make_pair(ip,hwaddr));
+                        *logger << " at " << hwaddr;
+                        break;
+                    }
+                }
+                *logger << "\n";
             }
         } catch(...)
         {
@@ -232,7 +251,7 @@ void SNMPQuery::fetchArpTable(Logger *logger, SNMPConnection *connection) throw(
             delete c;
         // If program was stopped, show this, rather
         // than error.
-        check_stop();
+        CHECK_STOP
         throw;
     }
     if(!connection)
@@ -241,13 +260,13 @@ void SNMPQuery::fetchArpTable(Logger *logger, SNMPConnection *connection) throw(
 }
 
 /**
- * Fetches routing table. If interfaces have been already fetched
+ * Fetches routing table. If interfaces have been already found
  * it will contain also reference to interface object associated
  * with the route.
  */
-void SNMPQuery::fetchRoutingTable(Logger *logger, SNMPConnection *connection) throw(FWException)
+void SNMPQuery::fetchRoutingTable(Logger *logger,SyncFlag *stop_program, SNMPConnection *connection) throw(FWException)
 {
-    *logger << "Fetching routing table\n"; 
+    *logger << "Routing table\n"; 
     routes.clear();
     
     SNMPConnection *c;
@@ -262,7 +281,7 @@ void SNMPQuery::fetchRoutingTable(Logger *logger, SNMPConnection *connection) th
         if(hostname.empty())
             throw FWException("No SNMP hostname specified");
         
-        check_stop();
+        CHECK_STOP
         c=new SNMPConnection(hostname, community);
         *logger << "Connecting to " << hostname << "\n";
         try
@@ -272,10 +291,10 @@ void SNMPQuery::fetchRoutingTable(Logger *logger, SNMPConnection *connection) th
         {
             // If program was stopped, show this, rather
             // than error.
-            check_stop();
+            CHECK_STOP
             throw;
         }
-        check_stop();
+        CHECK_STOP
     }
     
     vector<SNMPVariable*> v;
@@ -283,24 +302,28 @@ void SNMPQuery::fetchRoutingTable(Logger *logger, SNMPConnection *connection) th
     {
         // Get Indices and destinations
         
-        check_stop();
+        CHECK_STOP
         multimap<string, SNMPVariable* > w=c->walk(SNMP_ROUTE_DST_TABLE);
         try
         {
-            check_stop();
+            CHECK_STOP
             for(multimap<string, SNMPVariable* >::iterator j = w.begin(); j!=w.end(); ++j)
             {
-                check_stop();
+                CHECK_STOP
                 if((*j).second->type!=SNMPVariable::snmp_ipaddr)
                 {
                     *logger << "unexpected result type in '" << 
-                    SNMP_ROUTE_DST_TABLE << "'table. Skipping it.\n";
+                    SNMP_ROUTE_DST_TABLE << "' table. Skipping it.\n";
                     continue;
                 }
                 IPAddress dst = dynamic_cast<SNMPVariable_IPaddr*>((*j).second)->getIPAddressValue();
-                string rname=(*j).first.substr(strlen(SNMP_ROUTE_DST_TABLE)+1);
+                string rname;
+		if ((*j).first.substr(0,strlen(RFC1213_MIB_PREFIX_IPROUTEDEST_PREFIX))==RFC1213_MIB_PREFIX_IPROUTEDEST_PREFIX)
+		  rname=(*j).first.substr(strlen(RFC1213_MIB_PREFIX_IPROUTEDEST_PREFIX)+1);
+		else
+		  rname=(*j).first.substr(strlen(SNMP_ROUTE_DST_TABLE)+1);
 
-                check_stop();
+                CHECK_STOP
                 v=c->get(string(SNMP_ROUTE_NM_TABLE)+"."+rname);
                 if(v[0]->type!=SNMPVariable::snmp_ipaddr)
                 {
@@ -312,12 +335,12 @@ void SNMPQuery::fetchRoutingTable(Logger *logger, SNMPConnection *connection) th
                 Netmask nm = dynamic_cast<SNMPVariable_IPaddr*>(v[0])->getNetmaskValue();
                 SNMPVariable::freeVarList(v);
 
-                check_stop();
+                CHECK_STOP
                 v=c->get(string(SNMP_ROUTE_TYPE_TABLE)+"."+rname);
                 long type = SNMPVariable::varList2Int(v);
                 SNMPVariable::freeVarList(v);
 
-                check_stop();
+                CHECK_STOP
                 v=c->get(string(SNMP_ROUTE_GW_TABLE)+"."+rname);
                 if(v[0]->type!=SNMPVariable::snmp_ipaddr)
                 {
@@ -329,11 +352,11 @@ void SNMPQuery::fetchRoutingTable(Logger *logger, SNMPConnection *connection) th
                 IPAddress gw = dynamic_cast<SNMPVariable_IPaddr*>(v[0])->getIPAddressValue();
                 SNMPVariable::freeVarList(v);
 
-                check_stop();
+                CHECK_STOP
                 v=c->get(string(SNMP_ROUTE_IF_TABLE)+"."+rname);
                 int ifindex = SNMPVariable::varList2Int(v);
                 SNMPVariable::freeVarList(v);
-                
+ 
                 map<int,Interface>::iterator ici=interfaces.find(ifindex);
                 
                 Interface *route_intf;
@@ -350,10 +373,23 @@ void SNMPQuery::fetchRoutingTable(Logger *logger, SNMPConnection *connection) th
                     
                     route_intf=&(*ici).second;
                 }
+                bool isdef= type!=SNMP_DIRECT_ROUTE &&
+                    nm.getLength() == 0 && dst == IPAddress("0.0.0.0");
+
+		if (isdef && route_intf) route_intf->setExt(true);
+
                 IPRoute rt(dst, nm, gw, route_intf, type==SNMP_DIRECT_ROUTE);
                 routes.push_back(rt);
-		if (route_intf)
-                    route_intf->setExt(isDefault(rt));
+
+                *logger << "route: " 
+                        << rt.getDestination().toString() 
+                        << "/" << rt.getNetmask().getLength() 
+                        << "  gw " 
+                        << rt.getGateway().toString()
+                        << "  "
+                        << string( (route_intf)?rt.getInterface()->getName():"(interface not found)")
+                        << string( (route_intf && rt.getInterface()->isExt())?"(Ext)":"")
+                        << "\n";
             }
         } catch(...)
         {
@@ -372,7 +408,7 @@ void SNMPQuery::fetchRoutingTable(Logger *logger, SNMPConnection *connection) th
             delete c;
         // If program was stopped, show this, rather
         // than error.
-        check_stop();
+        CHECK_STOP
         throw;
     }
     if(!connection)
@@ -383,11 +419,11 @@ void SNMPQuery::fetchRoutingTable(Logger *logger, SNMPConnection *connection) th
 bool SNMPQuery::isDefault(const IPRoute &r) const
 {
     return !r.isDirect() &&
-    r.getNetmask() == Netmask("0.0.0.0") &&
+    r.getNetmask().getLength() == 0 &&
     r.getDestination() == IPAddress("0.0.0.0");
 }
 
-void SNMPQuery::fetchInterfaces(Logger *logger, SNMPConnection *connection) throw(FWException)
+void SNMPQuery::fetchInterfaces(Logger *logger,SyncFlag *stop_program, SNMPConnection *connection) throw(FWException)
 {
     interfaces.clear();
     
@@ -403,7 +439,7 @@ void SNMPQuery::fetchInterfaces(Logger *logger, SNMPConnection *connection) thro
         if(hostname.empty())
             throw FWException("No SNMP hostname specified");
         
-        check_stop();
+        CHECK_STOP
         c=new SNMPConnection(hostname, community);
         *logger << "Connecting to " << hostname << "\n";
         try
@@ -413,48 +449,64 @@ void SNMPQuery::fetchInterfaces(Logger *logger, SNMPConnection *connection) thro
         {
             // If program was stopped, show this, rather
             // than error.
-            check_stop();
+            CHECK_STOP
             throw;
         }
-        check_stop();
+        CHECK_STOP
     }
     
     multimap<string, SNMPVariable* > w;
     vector<SNMPVariable*>            v;
     try
     {
-        *logger << "Fetching interfaces address indices.\n";
-        
-        map<long, string> addr;
+        *logger << "Index of interface addresses.\n";
+
+        map<long, list<string> > addr;
         w=c->walk(SNMP_ADDR_INDEX_TABLE);
-        check_stop();
+        CHECK_STOP
+
         for(multimap<string, SNMPVariable* >::iterator j = w.begin(); j!=w.end(); ++j)
         {
-            check_stop();
+            CHECK_STOP
             if((*j).second->type!=SNMPVariable::snmp_int)
             {
                 *logger << "unexpected result type in '" << 
-                SNMP_ADDR_INDEX_TABLE << "'table. Skipping it.\n";
+                SNMP_ADDR_INDEX_TABLE << "' table. Skipping it.\n";
                 continue;
             }
             long ind = SNMPVariable::var2Int((*j).second);
-            string iname=(*j).first.substr(strlen(SNMP_ADDR_INDEX_TABLE)+1);
-            addr[ind]=iname;
+	    string iname;
+	    if ((*j).first.substr(0,strlen(IP_MIB_ENTIFINDEX_PREFIX))==IP_MIB_ENTIFINDEX_PREFIX)
+	      iname=(*j).first.substr(strlen(IP_MIB_ENTIFINDEX_PREFIX)+1);
+	    else
+              iname=(*j).first.substr(strlen(SNMP_ADDR_INDEX_TABLE)+1);
+
+/*
+ * Example (two interfaces, first has two addresses):
+ *
+ * ip.ipAddrTable.ipAddrEntry.ipAdEntIfIndex.10.3.14.201 = 1
+ * ip.ipAddrTable.ipAddrEntry.ipAdEntIfIndex.192.168.100.100 = 1
+ * ip.ipAddrTable.ipAddrEntry.ipAdEntIfIndex.222.22.22.222 = 2
+ *
+ * at this point iname is watever text we get after
+ * "ip.ipAddrTable.ipAddrEntry.ipAdEntIfIndex." , ind is an index
+ */
+            addr[ind].push_back(iname);
+            *logger << "interface #" << ind << ": " << iname << "\n";
         }
 
         for(multimap<string, SNMPVariable* >::iterator j = w.begin(); j!=w.end(); ++j)
             delete (*j).second;
         w.clear();
 
-
-        check_stop();
-        *logger << "Fetching full interfaces information.\n";
+        CHECK_STOP
+        *logger << "Full interfaces info.\n";
 
         w=c->walk(SNMP_INTERFACE_INDEX);
-        check_stop();
+        CHECK_STOP
         for(multimap<string, SNMPVariable* >::iterator j = w.begin(); j!=w.end(); ++j)
         {
-            check_stop();
+            CHECK_STOP
 
             // Get index
             long ifindex = SNMPVariable::var2Int((*j).second);
@@ -467,7 +519,7 @@ void SNMPQuery::fetchInterfaces(Logger *logger, SNMPConnection *connection) thro
                 throw FWException(string("Unexpected response length for variable: ")+str);
             int astatus = SNMPVariable::varList2Int(v);
             SNMPVariable::freeVarList(v);
-            check_stop();
+            CHECK_STOP
 
             if(!astatus)
             {
@@ -482,9 +534,18 @@ void SNMPQuery::fetchInterfaces(Logger *logger, SNMPConnection *connection) thro
                 throw FWException(string("Unexpected response length for variable: ")+str);
             int ostatus = SNMPVariable::varList2Int(v);
             SNMPVariable::freeVarList(v);
-            check_stop();
-            
-            string aprefix;
+            CHECK_STOP
+
+
+/* gather all information for interface ifindex and create Interface object */
+
+            list<string> &addlist=addr[ifindex];
+            if (addlist.empty())
+            {
+                *logger << "Interface #" << ifindex << " has no IP address.\n";
+//                continue;
+            }
+/*
             map<long, string>::const_iterator ici=addr.find(ifindex);
             if(ici==addr.end())
             {
@@ -494,13 +555,13 @@ void SNMPQuery::fetchInterfaces(Logger *logger, SNMPConnection *connection) thro
             {
                 aprefix=(*ici).second;
             }
-
+*/
             // Get desriptions
             sprintf(str,"%s.%ld", SNMP_INTERFACES_DESCR , ifindex);
             v=c->get(str);
             string descr = SNMPVariable::varList2String(v);
             SNMPVariable::freeVarList(v);
-            check_stop();
+            CHECK_STOP
 
             // Get physical address
             sprintf(str,"%s.%ld", SNMP_INTERFACES_PHYSA , ifindex);
@@ -511,7 +572,7 @@ void SNMPQuery::fetchInterfaces(Logger *logger, SNMPConnection *connection) thro
                 throw FWException(string("Unexpected response type for: ")+str);
             string physa = dynamic_cast<SNMPVariable_String*>(v[0])->toHexString();
             SNMPVariable::freeVarList(v);
-            check_stop();
+            CHECK_STOP
 
             // Get type
             sprintf(str,"%s.%ld", SNMP_INTERFACES_TYPE , ifindex);
@@ -520,41 +581,57 @@ void SNMPQuery::fetchInterfaces(Logger *logger, SNMPConnection *connection) thro
                 throw FWException(string("Unexpected response length for variable: ")+str);                
             int itype = SNMPVariable::varList2Int(v);
             SNMPVariable::freeVarList(v);
-            check_stop();
+            CHECK_STOP
 
-            // get netmasks
-            check_stop();
-            v=c->get(string(SNMP_NMASK_TABLE)+"."+aprefix);
-            if(v.size()!=1)
-                throw FWException("Can't get netmask");
-            if(v[0]->type!=SNMPVariable::snmp_ipaddr)
-                throw FWException("Wrong return type for netmask");
-            string nm = dynamic_cast<SNMPVariable_IPaddr*>(v[0])->getNetmaskValue().toString();
-            SNMPVariable::freeVarList(v);
-            check_stop();
+            interfaces[ifindex]=Interface();
+            interfaces[ifindex].setName(descr);
+            interfaces[ifindex].setOStatus(ostatus);
+            interfaces[ifindex].setInterfaceType(itype);
+            interfaces[ifindex].setExt(false);
+
+            if (physa!="")
+                interfaces[ifindex].setPhysicalAddress(physa);
+
+            for (list<string>::iterator ali=addlist.begin(); ali!=addlist.end(); ali++)
+            {
+                string aprefix= *ali;
+
+                // get netmasks
+                CHECK_STOP
+                v=c->get(string(SNMP_NMASK_TABLE)+"."+aprefix);
+                if(v.size()!=1)
+                    throw FWException("Can't get netmask");
+                if(v[0]->type!=SNMPVariable::snmp_ipaddr)
+                    throw FWException("Wrong return type for netmask");
+                string nm = dynamic_cast<SNMPVariable_IPaddr*>(v[0])->getNetmaskValue().toString();
+                SNMPVariable::freeVarList(v);
+                CHECK_STOP
             
-            // get addresss
-            v=c->get(string(SNMP_ADDR_TABLE)+"."+aprefix);
-            if(v.size()!=1)
-                throw FWException("Can't get IP address");
-            if(v[0]->type!=SNMPVariable::snmp_ipaddr)
-                throw FWException("Wrong return type for IP address");
-            string ad = dynamic_cast<SNMPVariable_IPaddr*>(v[0])->getIPAddressValue().toString();
-            SNMPVariable::freeVarList(v);
-            check_stop();
+                // get addresss
+                v=c->get(string(SNMP_ADDR_TABLE)+"."+aprefix);
+                if(v.size()!=1)
+                    throw FWException("Can't get IP address");
+                if(v[0]->type!=SNMPVariable::snmp_ipaddr)
+                    throw FWException("Wrong return type for IP address");
+                string ad = dynamic_cast<SNMPVariable_IPaddr*>(v[0])->getIPAddressValue().toString();
+                SNMPVariable::freeVarList(v);
+                CHECK_STOP
 
-            v=c->get(string(SNMP_BCAST_TABLE)+"."+aprefix);
-            if(v.size()!=1)
-                throw FWException("Can't get broadcast address");
-            long bcast_bits = SNMPVariable::varList2Int(v);
-            SNMPVariable::freeVarList(v);
-            check_stop();
+                v=c->get(string(SNMP_BCAST_TABLE)+"."+aprefix);
+                if(v.size()!=1)
+                    throw FWException("Can't get broadcast address");
+//                long bcast_bits = SNMPVariable::varList2Int(v);
+                SNMPVariable::freeVarList(v);
+                CHECK_STOP
 
-            interfaces[ifindex]=Interface(descr, ad, nm, bcast_bits, itype, ostatus,  false, false, physa);
-#ifdef SNMP_CRAWLER_DEBUG
-            *logger << "Adding interface # " << ifindex << "\n";
-#endif
+                IPv4 *ipaddr=interfaces[ifindex].addIPv4();
+                ipaddr->setAddress(ad);
+                ipaddr->setNetmask(nm);
 
+                *logger << "Adding interface #" << ifindex 
+                        << ": " << ad << "/" << nm
+                        << "\n";
+            }
         } // index walk.
         for(multimap<string, SNMPVariable* >::iterator j = w.begin(); j!=w.end(); ++j)
             delete (*j).second;
@@ -573,7 +650,8 @@ void SNMPQuery::fetchInterfaces(Logger *logger, SNMPConnection *connection) thro
             delete c;
         // If program was stopped, show this, rather
         // than error.
-        check_stop();
+        CHECK_STOP
+
         throw;
     }
     if(!connection)
@@ -581,7 +659,7 @@ void SNMPQuery::fetchInterfaces(Logger *logger, SNMPConnection *connection) thro
     *logger << "Done fetching interfaces\n";
 }
 
-void SNMPQuery::fetchSysInfo(Logger *logger, SNMPConnection *connection) throw(FWException)
+void SNMPQuery::fetchSysInfo(Logger *logger,SyncFlag *stop_program, SNMPConnection *connection) throw(FWException)
 {
     descr     = "" ;
     contact   = "" ;
@@ -596,17 +674,17 @@ void SNMPQuery::fetchSysInfo(Logger *logger, SNMPConnection *connection) throw(F
     {
         if(community.empty())
         {
-            check_stop();
+            CHECK_STOP
             throw FWException("No SNMP community specified");
         }
         
         if(hostname.empty())
         {
-            check_stop();
+            CHECK_STOP
             throw FWException("No SNMP hostname specified");
         }
         
-        check_stop();
+        CHECK_STOP
         c=new SNMPConnection(hostname, community);
         *logger << "Connecting to " << hostname << "\n";
         try
@@ -616,10 +694,10 @@ void SNMPQuery::fetchSysInfo(Logger *logger, SNMPConnection *connection) throw(F
         {
             // If program was stopped, show this, rather
             // than error.
-            check_stop();
+            CHECK_STOP
             throw;
         }
-        check_stop();
+        CHECK_STOP
     }
     
     try
@@ -630,32 +708,32 @@ void SNMPQuery::fetchSysInfo(Logger *logger, SNMPConnection *connection) throw(F
         v=c->get(SNMP_SYSNAME);
         sysname    = SNMPVariable::varList2String(v);
         SNMPVariable::freeVarList(v);
-        check_stop();
+        CHECK_STOP
 
         *logger << "Getting Description\n";
         v=c->get(SNMP_SYSDESCR);
         descr    = SNMPVariable::varList2String(v);
         SNMPVariable::freeVarList(v);
-        check_stop();
+        CHECK_STOP
         
         *logger << "Getting Location\n";
         v=c->get(SNMP_LOCATION);
         location = SNMPVariable::varList2String(v);
         SNMPVariable::freeVarList(v);
-        check_stop();
+        CHECK_STOP
         
         *logger << "Getting Contact Info\n";
         v=c->get(SNMP_CONTACT);
         contact  = SNMPVariable::varList2String(v);
         SNMPVariable::freeVarList(v);
-        check_stop();
+        CHECK_STOP
     } catch(...)
     {
         if(!connection)
             delete c;
         // If program was stopped, show this, rather
         // than error.
-        check_stop();
+        CHECK_STOP
         throw;
     }
     if(!connection)
@@ -799,11 +877,11 @@ multimap<string, SNMPVariable* > SNMPConnection::walk(const string &variable) th
                         continue;
                     }
                     char n[MAX_NAME_LEN];
-#ifdef OVERRUN_SAFE_UCD_SNMP
-                    sprint_objid (binit(NULL, n, sizeof(n)), vars->name, vars->name_length);
+#ifdef HAVE_SNPRINT_OBJID
+                    snprint_objid (n, sizeof(n), vars->name, vars->name_length);
 #else
                     sprint_objid (n, vars->name, vars->name_length);
-#endif                    
+#endif
                     res.insert(make_pair(string(n),SNMPVariable::create(vars)));
                     if ((vars->type != SNMP_ENDOFMIBVIEW) &&
                         (vars->type != SNMP_NOSUCHOBJECT) &&
@@ -1134,7 +1212,8 @@ bool SNMPCrawler::isvirtual(const IPAddress &addr, const string &pa) const
         {
             try
             {
-                if(pa == (*j).second.getPhysicalAddress() && addr != (*j).second.getIPAddress())
+                physAddress *paddr=(*j).second.getPhysicalAddress();
+                if(paddr!=NULL && pa == paddr->getPhysAddress() && addr != (*j).second.getIPAddress())
                     return true;
             } catch(FWException &ex)
             {
@@ -1156,7 +1235,7 @@ bool SNMPCrawler::point2point(const Interface *intf) const
 {
     if(intf)
     {
-        int itype=intf->getSNMPType();
+        int itype=intf->getInterfaceType();
         for(unsigned int i=0;i<(sizeof(PTP_INTERFACE_TYPES)/sizeof(int));i++)
             if(PTP_INTERFACE_TYPES[i]==itype)
                 return true;
@@ -1184,13 +1263,13 @@ bool SNMPCrawler::special(const IPNetwork &n) const
 }
 
 //TODO: multiple threads (via pool).
-void SNMPCrawler::run_impl(Logger *logger) throw(FWException)
+void SNMPCrawler::run_impl(Logger *logger,SyncFlag *stop_program) throw(FWException)
 {
     time_t now=time(NULL);
-    *logger << "Starting SNMPCrawler at " << asctime(localtime(&now)) << ". Seed host: " << (*(queue.begin())).first.toString() << "\n";
+    *logger << "SNMPCrawler started at " << asctime(localtime(&now)) << ". Seed host: " << (*(queue.begin())).first.toString() << "\n";
     SNMP_discover_query q;
     do {
-        check_stop();
+        CHECK_STOP
         map<IPAddress,string>::iterator i=queue.begin();
         if(i==queue.end())
         {
@@ -1200,7 +1279,7 @@ void SNMPCrawler::run_impl(Logger *logger) throw(FWException)
         string task_phys_address = (*i).second ;
         queue.erase(i);
         
-        check_stop();
+        CHECK_STOP
 
         *logger << "Processing " << task.toString() << "\n";
 
@@ -1212,29 +1291,36 @@ void SNMPCrawler::run_impl(Logger *logger) throw(FWException)
                snmp_timeout
         );
         
-        check_stop();
+        CHECK_STOP
 
         CrawlerFind *res=new CrawlerFind();
         found[task] = *res; delete res;
         found[task].found_phys_addr = task_phys_address ;
         try
         {
-            q.fetchArpTable(logger);
+            q.fetchArpTable(logger,stop_program);
             found[task].have_snmpd = true;
         } catch(const FWException &ex)
         {
             // fetch failed
-            *logger << "Failed to fetch ARP table from " << task.toString() << "\n";
+            *logger << "Failed to fetch ARP table from " << task.toString();
+            if (!ex.toString().empty()) *logger << " : " << ex.toString();
+            *logger << "\n";
+            continue;
+        } catch (std::string s) {
+            *logger << "Failed to fetch ARP table from " << task.toString();
+            *logger << " : " << s;
+            *logger << "\n";
             continue;
         }
-        check_stop();
+        CHECK_STOP
         map<IPAddress, string> at=q.getArtpTable();    
-        *logger << "Fetched " << long(at.size()) << " entries\n";
+        *logger << "Got " << long(at.size()) << " entries\n";
 
         int qplus=0, rplus=0, dplus=0;
         for(map<IPAddress, string>::iterator j=at.begin();j!=at.end();++j)
         {
-            check_stop();
+            CHECK_STOP
 
             IPAddress  c = (*j).first;
             string    pa = (*j).second;
@@ -1261,7 +1347,7 @@ void SNMPCrawler::run_impl(Logger *logger) throw(FWException)
 
         try
         {
-            q.fetchInterfaces(logger);
+            q.fetchInterfaces(logger,stop_program);
             found[task].have_snmpd = true;
 
             map<int, Interface> intf = q.getInterfaces();
@@ -1280,7 +1366,8 @@ void SNMPCrawler::run_impl(Logger *logger) throw(FWException)
                     networks.insert(net);
                 }
             }
-        } catch(const FWException &ex)
+//        } catch(const FWException &ex)
+        } catch(FWException &ex)
         {
             // fetch failed
             *logger << "Failed to fetch list of interfaces from " << task.toString() << "\n";
@@ -1288,7 +1375,7 @@ void SNMPCrawler::run_impl(Logger *logger) throw(FWException)
 
         try
         {
-            q.fetchSysInfo(logger);
+            q.fetchSysInfo(logger,stop_program);
             found[task].have_snmpd = true;
 
             found[task].sysname  = q.getSysname  ();
@@ -1303,7 +1390,7 @@ void SNMPCrawler::run_impl(Logger *logger) throw(FWException)
 
         try
         {
-            q.fetchRoutingTable(logger);
+            q.fetchRoutingTable(logger,stop_program);
             found[task].have_snmpd = true;
             
             vector<IPRoute> routes = q.getRoutes();
@@ -1342,7 +1429,7 @@ void SNMPCrawler::run_impl(Logger *logger) throw(FWException)
                     // since this route will be ignored because of it.
                     //
                     // 2. From interfaces which are UP, we select first one
-                    // which have Point-to-Point attribute enabled.
+                    // with Point-to-Point attribute enabled.
                     //
                     for(set<Interface>::const_iterator k=gi.begin(); k!=gi.end(); ++k)
                     {
@@ -1373,11 +1460,11 @@ void SNMPCrawler::run_impl(Logger *logger) throw(FWException)
                     }
                     if(have_intf)
                         *logger << "Guessed that network " << net << 
-                        " is using interface with IP " << string(intf.getIPAddress()) << "\n";
+                        " is using interface " << intf.getName() << "\n";
                 }
 
-                // If route associated with interface which is down
-                // it will be ignored.
+                // If route is associated with an interface which is down,
+                // ignore it.
                 if(have_intf && !intf.isUp())
                 {
                     *logger << "Skipping route for network " << net << 
@@ -1459,16 +1546,16 @@ void SNMPCrawler::run_impl(Logger *logger) throw(FWException)
     } while(recursive);
 
     if(skip_virtual)
-        remove_virtual(logger);
+        remove_virtual(logger,stop_program);
     
     if(do_dns)
-        bacresolve_results(logger);
+        bacresolve_results(logger,stop_program);
     
     now=time(NULL);
     *logger << "SNMPCrawler - done at " << asctime(localtime(&now)) << "\n";
 }
 
-void SNMPCrawler::remove_virtual(Logger *logger) throw(FWException)
+void SNMPCrawler::remove_virtual(Logger *logger,SyncFlag *stop_program) throw(FWException)
 {
     *logger << "Removing virtual IPs.\n";
     for(map<IPAddress, CrawlerFind>::iterator j=found.begin(); j!=found.end(); ++j)
@@ -1476,9 +1563,9 @@ void SNMPCrawler::remove_virtual(Logger *logger) throw(FWException)
             found.erase(j);
 }
 
-void SNMPCrawler::bacresolve_results(Logger *logger) throw(FWException)
+void SNMPCrawler::bacresolve_results(Logger *logger,SyncFlag *stop_program) throw(FWException)
 {
-    *logger << "Resolving found hosts.\n";
+    *logger << "Resolving names\n";
         
     set<IPAddress> resolv_set;
     for(map<IPAddress, CrawlerFind>::iterator j=found.begin(); j!=found.end(); ++j)
@@ -1489,12 +1576,13 @@ void SNMPCrawler::bacresolve_results(Logger *logger) throw(FWException)
         
     try
     {
-        backresq.run_impl(logger);
+        backresq.run_impl(logger,stop_program);
     } catch(const FWException &ex)
     {
-        *logger << "Resolving reported errors.\n";
+        *logger << "DNS resolver reported errors.\n";
     }
-        
+    *logger << "Finished.\n";
+
     map<IPAddress, HostEnt> resolv_res=backresq.getResult();
     for(map<IPAddress, HostEnt>::iterator j=resolv_res.begin(); j!=resolv_res.end(); ++j)
     {
@@ -1525,32 +1613,32 @@ CrawlerFind::~CrawlerFind()
 {
 }
 
-void SNMP_interface_query::run_impl(Logger *logger) throw(FWException)
+void SNMP_interface_query::run_impl(Logger *logger,SyncFlag *stop_program) throw(FWException)
 {
-    fetchInterfaces(logger);
+    fetchInterfaces(logger,stop_program);
     
     // We try to fetch routing table, to find
     // which interface is "external".
     // We do not fail if this query does not succeed.
     try
     {
-        fetchRoutingTable(logger);
+        fetchRoutingTable(logger,stop_program);
     } catch(FWException &ex)
     {
         *logger << "Error fetching routing table, external interface will not be detected.\n";
     }
 }
 
-void SNMP_sysdesc_query::run_impl(Logger *logger) throw(FWException)
+void SNMP_sysdesc_query::run_impl(Logger *logger,SyncFlag *stop_program) throw(FWException)
 {
-    fetchSysInfo(logger);
+    fetchSysInfo(logger,stop_program);
 }
 
-void SNMP_discover_query::run_impl(Logger *logger) throw(FWException)
+void SNMP_discover_query::run_impl(Logger *logger,SyncFlag *stop_program) throw(FWException)
 {
-    fetchArpTable(logger);
+    fetchArpTable(logger,stop_program);
     if(fetch_inerfaces)
-        fetchInterfaces(logger);
+        fetchInterfaces(logger,stop_program);
 }
 
 
